# -*- coding: utf-8 -*-
"""Base classes to define and create word representations.

This module provides the abstract base class `Representation` with two implementations and the main function
`create_representation` to construct one..
"""
import abc
import gzip
import logging
import re

import numpy as np
from numpy import dtype
import pandas as pd

import mangoes.evaluation.analogy
import mangoes.utils
import mangoes.utils.arrays
import mangoes.utils.persist
import mangoes.utils.exceptions
from mangoes.constants import ENCODING

_logger = logging.getLogger(__name__)


class Representation(abc.ABC):
    """Abstract base class for a Representation.

    Parameters
    ----------
    words: mangoes.Vocabulary
        words represented as vectors (rows of the matrix)
    matrix: mangoes.utils.arrays.Matrix
        vectors representing words

    """

    def __init__(self, words, matrix, hyperparameters=None):
        self._params = {}
        self.words = words
        self._params["words"] = self.words.params
        self.matrix = mangoes.utils.arrays.Matrix.factory(matrix)
        if hyperparameters:
            self._params.update(hyperparameters)

    def __getitem__(self, word):
        return self.matrix[self.words.index(word)]

    @property
    def shape(self):
        return self.matrix.shape

    @property
    def params(self):
        """Dict of the parameters used to build this matrix, if available"""
        return self._params

    def pprint(self, display=print):
        """Pretty print the matrix with labels for rows and columns"""
        pd.set_option('display.width', None)
        display(self.to_df())

    @abc.abstractmethod
    def to_df(self):
        """Returns a pandas DataFrame representation of this matrix"""
        pass

    @classmethod
    def load(cls, path):
        """Load a Representation

        According to files found in the path, the Representation is either a CountBasedRepresentation object
        or an Embeddings object.

        Parameters
        -----------
        path: str
            Path to a folder or an archive
        """
        try:
            return mangoes.CountBasedRepresentation.load(path)
        except FileNotFoundError:
            return mangoes.Embeddings.load(path)

    def save(self, path):
        """Save the Representation

        Parameters
        ----------
        path: str
            Path to a folder or an archive. Will be created if doesn't exist.
        """
        mangoes.utils.persist.save(self, self._attributes_to_persist, path, self._params)

    def distance(self, word1, word2, metric="cosine", **kwargs):
        """Returns the distance between two words

        This function relies on the `sklearn.metrics.pairwise_distances` module so you can use any distance
        available in it.

        Parameters
        ----------
        word1: str
            First word
        word2: str
            Second word
        metric: str
            The metric to use
        **kwargs : optional keyword parameters
            Any further parameters are passed directly to the distance function.
        """
        return self.pairwise_distances([word1], [word2], metric, **kwargs)[0, 0]

    def pairwise_distances(self, words, other_words=None, metric="cosine", **kwargs):
        """ Compute the distance matrix from one or two list(s) of words.

        This method takes a list of words, and returns the matrix of the distances between them.
        Or, if `other_words` is given (default is None), then the returned matrix is the pairwise
        distance between the words from `words` and the ones from `other_words`.

        This function relies on the `sklearn.metrics.pairwise_distances` module so you can use any distance
        available in it.

        Valid values for metric are:
        - From scikit-learn: ['cityblock', 'cosine', 'euclidean', 'l1', 'l2', 'manhattan'].
          These metrics support sparse matrix inputs.
        - From scipy.spatial.distance: ['braycurtis', 'canberra', 'chebyshev', 'correlation', 'dice', 'hamming',
          'jaccard', 'kulsinski', 'mahalanobis', 'matching', 'minkowski', 'rogerstanimoto', 'russellrao', 'seuclidean',
          'sokalmichener', 'sokalsneath', 'sqeuclidean', 'yule']

        See the documentation for sklearn.metrics.pairwise_distances for details on these metrics.

        Parameters
        ----------
        words : list of str
            List of words
        other_words: list of str, optional
            An optional second list.
        metric : string, or callable
            The metric to use when calculating distance between words vectors.
        **kwargs : optional keyword parameters
            Any further parameters are passed directly to the distance function.

        Returns
        -------
        array [len(words), len(words)] or [len(words), len(other_words)]
            A distance matrix D such that D_{i, j} is the distance between the ith and jth words of the `words` list,
            if `other_words` is None.
            If `other_words` is not None, then D_{i, j} is the distance between the ith word from `words` and the
            jth word from `other_words`.
        """
        vectors = mangoes.utils.arrays.Matrix.factory(self.matrix[self.words.indices(words), :])
        other_vectors = self.matrix[self.words.indices(other_words), :] if other_words is not None else None
        return vectors.pairwise_distances(other_vectors, metric=metric, **kwargs)

    def word_mover_distance(self, sentence1, sentence2, stopwords=None, metric='euclidean', emd=None):
        return mangoes.utils.metrics.word_mover_distance(self, sentence1, sentence2,
                                                         stopwords=stopwords, metric=metric, emd=emd)

    def get_closest_words(self, word, nb=10, metric="cosine"):
        """Returns the closest words and their distances from the given word

        Attributes
        ----------
        word: str, Token or vector
            a word or a vector representing a word
        nb: int
            the number of neighbors to return
        metric: str or callable
            the metric to use (see pairwise_distances method)

        Returns
        -------
        list
            list of tuples (word, distance) sorted by distance
        """
        if isinstance(word, str) or isinstance(word, tuple):
            # word parameter is a word (not a vector)
            distances = self.matrix.pairwise_distances(self[word], metric=metric).flatten()
            first, last = 1, nb + 1
        else:
            # word parameter is a vector
            distances = self.matrix.pairwise_distances(word, metric=metric).flatten()
            first, last = 0, nb

        sorted_indices = np.argsort(distances)
        return [(self.words[i], distances[i]) for i in sorted_indices[first:last]]

    def analogy(self, question):
        return mangoes.evaluation.analogy.Evaluator(self).predict(question)


class Embeddings(Representation):
    """Base class for a Word Embedding.

    Parameters
    ----------
    words: mangoes.Vocabulary
        words represented as vectors (rows of the matrix)
    matrix: mangoes.utils.arrays.Matrix
        vectors representing words

    """
    _attributes_to_persist = [('words', mangoes.Vocabulary), ('matrix', mangoes.utils.arrays.Matrix)]

    def __init__(self, words, matrix, hyperparameters=None):
        super().__init__(words, matrix, hyperparameters)

    def __len__(self):
        return len(self.words)

    def to_df(self):
        """Returns a pandas DataFrame representation of this matrix with labels for rows"""
        return pd.DataFrame(self.matrix, index=self.words, dtype=float)

    @classmethod
    def load(cls, path):
        """Load an Embeddings

        This loader expects to find in path :

        * a file named 'matrix.npy' or 'matrix.npz' for the matrix
        * a text files 'words.txt' with the words represented as vectors in the matrix

        Parameters
        -----------
        path: str
            Path to a folder or an archive
        """
        return mangoes.utils.persist.load(cls, cls._attributes_to_persist, path)

    @classmethod
    def load_from_gensim(cls, path):
        """Load an Embeddings from gensim

        Parameters
        ----------
        path: str


        Returns
        -------

        """
        import gensim
        model = gensim.models.KeyedVectors.load(path)
        vocabulary = mangoes.Vocabulary(sorted(model.wv.vocab, key=lambda w: model.wv.vocab[w].index))
        matrix = mangoes.utils.arrays.Matrix.factory(np.array(model.wv.vectors, dtype=float))
        return mangoes.Embeddings(vocabulary, matrix)

    def save_as_text_file(self, file_path, compress=False, sep="\t"):
        """Save the embedding as a text file, with one word and its corresponding list of embedding
        values per line.

        Parameters
        ----------
        file_path: str
            path to the location where to store the Embeddings instance as a text file
        compress: boolean
            whether or not to compress the output text file (default = False).
            If True, it will be compressed using 'gz', and be named accordingly.
        sep: str
            the string that shall act as the delimiter between words and/or between numbers on a line.
            (default = '\t')

        """
        sep_invalid_values = [",", ".", ":"]
        if sep in sep_invalid_values:
            # TODO : check and explain which value are not accepted and why
            error_message = "Wrong 'sep' input: the following values are not accepted: {}. " \
                            "Default value '\t' is recommended.".format(str(sep_invalid_values))
            raise mangoes.utils.exceptions.NotAllowedValue(error_message)

        # Determine whether or not to write a compressed file
        compresslevel = 0
        _get_open_function = lambda: open(file_path, 'w', encoding=ENCODING)
        if compress or file_path.endswith(".gz"):
            compresslevel = 9
            if not file_path.endswith(".gz"):
                file_path = "{:s}{:s}".format(file_path, ".gz")
            _get_open_function = lambda: gzip.open(file_path, 'wt', compresslevel=compresslevel,
                                                   encoding=ENCODING)

        # Write embedding in file
        with _get_open_function() as f:
            if self.matrix.is_sparse:
                f.write("({:d}, {:d}), {dtype:s}\n".format(*self.matrix.shape,
                                                           dtype=repr(self.matrix.dtype)))
            for i, vector in enumerate(self.matrix):
                word = self.words.words[i]
                string_vector = self.matrix.__class__.format_vector(vector, sep)
                f.write("{:s}{:s}{:s}\n".format(word, sep, string_vector))

    @staticmethod
    def load_from_text_file(file_path, sep="\t"):
        """Load an embedding from a text file, where there is one word and its corresponding list
        of embedding values per line.

        The text file may be in a compressed format, such as '.gz'.

        Parameters
        ----------
        file_path: str
            path to the text file containing the Embeddings' instance's data
        sep: str
            the string that shall act as the delimiter between words and/or between numbers on a line.
            (default = '\t')

        Returns
        -------
        Embeddings
        """
        _get_open_function = lambda: open(file_path, 'r', encoding=ENCODING)
        if file_path.endswith(".gz"):
            _get_open_function = lambda: gzip.open(file_path, 'rt', encoding=ENCODING)

        with _get_open_function() as f:
            line = f.readline()
            match = re.match(r"\((\d+), (\d+)\), (dtype.*)", line)
            if match:
                nb_columns = match.group(2)
                data_type = eval(match.group(3))
                matrix, words_list = mangoes.utils.arrays.csrSparseMatrix.load_from_text_file(f, int(nb_columns),
                                                                                              data_type,
                                                                                              sep)
            else:
                f.seek(0)
                matrix, words_list = mangoes.utils.arrays.NumpyMatrix.load_from_text_file(f, sep)

        return Embeddings(mangoes.Vocabulary(words_list), matrix)

    @classmethod
    def load_from_pickle_files(cls, matrix_file_path, vocabulary_file_path=None):
        """Load an Embeddings instance from a matrix and vocabulary pickle file(s).

        Parameters
        ----------
        matrix_file_path: str
            path to the pickle file where is stored at least the matrix (if vocabulary_file_path is not None)
            and also the vocabulary (if vocabulary_file_path is None).
        vocabulary_file_path: str, optional (default=None)
            path to the pickle file, where the vocabulary is stored, if the matrix and the vocabulary are in
            separate files

        Returns
        -------
        Embeddings
        """
        import pickle
        if vocabulary_file_path is None:
            with open(matrix_file_path, "rb") as f:
                object1 = pickle.load(f)
                object2 = pickle.load(f)
                try:
                    matrix = mangoes.utils.arrays.Matrix.factory(object1)
                    words = object2
                except (TypeError, NotImplementedError):
                    matrix = mangoes.utils.arrays.Matrix.factory(object2)
                    words = object1
        else:
            with open(matrix_file_path, "rb") as f:
                matrix = pickle.load(f)
            matrix = mangoes.utils.arrays.Matrix.factory(matrix)
            with open(vocabulary_file_path, "rb") as f:
                words = pickle.load(f)
        return Embeddings(mangoes.Vocabulary(words), matrix)


class CountBasedRepresentation(Representation):
    """Base class for a cooccurrence count matrix.

    Parameters
    ----------
    words_vocabulary: mangoes.Vocabulary
        words represented as vectors (rows of the matrix)
    contexts_vocabulary: mangoes.Vocabulary
        words used as features (columns of the matrix)
    matrix
        numbers of cooccurrence

    """
    _attributes_to_persist = [('words', mangoes.Vocabulary),
                              ('contexts_words', mangoes.Vocabulary),
                              ('matrix', mangoes.utils.arrays.Matrix)]

    def __init__(self, words_vocabulary, contexts_vocabulary, matrix, hyperparameters=None):
        super().__init__(words_vocabulary, matrix, hyperparameters)
        self.contexts_words = contexts_vocabulary
        self._params["contexts_words"] = self.contexts_words.params

    def __iter__(self):
        return self.words.word_index.__iter__()

    def __eq__(self, other):
        # Equality between vocabularies
        if self.words != other.words:
            return False
        if self.contexts_words != other.contexts_words:
            return False
        # Equality between matrix
        return self.matrix == other.matrix

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_df(self):
        """Returns a pandas DataFrame representation of this matrix with labels for rows and columns"""
        if self.matrix.is_sparse:
            return pd.DataFrame.sparse.from_spmatrix(self.matrix,
                                                     index=self.words, columns=self.contexts_words)
        return pd.DataFrame(self.matrix,
                            index=self.words, columns=self.contexts_words,
                            dtype=int)

    @classmethod
    def load(cls, path):
        """Load a CooccurrenceCount

        This loader expects to find in path :

        * a file named 'matrix.npy' or 'matrix.npz' for the matrix
        * two text files 'words.txt' and 'contexts_words.txt' with the words used in rows and columns
          of the matrix, respectively

        Parameters
        -----------
        path: str
            Path to a folder or an archive
        """
        return mangoes.utils.persist.load(cls, cls._attributes_to_persist, path)


class Transformation:
    """Base callable class to define transformation to be applied to a Matrix

    See Also
    --------
    :func:`mangoes.create_representation`
    :mod:`mangoes.weighting`
    :mod:`mangoes.reduction`
    """

    def __init__(self):
        self._params = {"name": self.__class__.__name__}

    @abc.abstractmethod
    def __call__(self, matrix):
        """Apply the transformation and return the transformed matrix

        Parameters
        ----------
        matrix: mangoes.utils.arrays.Matrix

        Returns
        -------
        mangoes.utils.arrays.Matrix
        """
        pass

    @property
    def params(self):
        return self._params


def create_representation(source, weighting=None, reduction=None):
    """Create an Embeddings object from a CooccurrenceCount

    Apply the function(s) passed in weighting and reduction parameters and returns a mangoes.Representation.

    Examples
    --------
    >>>  embedding = mangoes.create_representation(cooccurrence_matrix,
    >>>                                           reduction=mangoes.reduction.pca(dimensions=50))

    Parameters
    ----------
    source: mangoes.CountBasedRepresentation
    weighting: mangoes.Transformation
        weighting function to apply to the source (see : `mangoes.weighting`)
    reduction: mangoes.Transformation
        reduction to apply to the (weighted) source matrix (see : `mangoes.reduction`)

    Returns
    --------
    Embeddings or CountBasedRepresentation

    See Also
    --------
    :mod:`mangoes.weighting`
    :mod:`mangoes.reduction`
    """
    matrix = source.matrix
    hyperparameters = {"source": source.params}

    if weighting:
        if not callable(weighting):
            msg = "`weighting` should be a callable, not a {}".format(type(weighting))
            raise mangoes.utils.exceptions.NotAllowedValue(msg)

        matrix = weighting(matrix)
        hyperparameters["weighting"] = weighting.params

    if reduction:
        if not callable(reduction):
            msg = "`reduction` should be a callable, not a {}".format(type(reduction))
            raise mangoes.utils.exceptions.NotAllowedValue(msg)

        matrix = reduction(matrix)
        hyperparameters["reduction"] = reduction.params
        return mangoes.base.Embeddings(source.words, matrix, hyperparameters)
    else:
        return mangoes.CountBasedRepresentation(source.words, source.contexts_words, matrix, hyperparameters)



import abc
import os.path
import re
import xml.etree.ElementTree
from collections import namedtuple
import string

import mangoes.utils

DIGIT_TOKEN = "0"
DIGIT_RE = r"\d+((,\d*)*(\.\d*)*)*"
PUNCTUATION_TRANSLATE_TABLE = str.maketrans('', '', string.punctuation)


def _compose(*args):
    if not args:
        def result(x):
            return x
    elif len(args) == 1:
        def result(x):
            return args[0](x)
    else:
        def result(x):
            return args[0](_compose(*args[1:])(x))

    return result


class SentenceGenerator:
    """Base class for sentences generators

    A sentence generator yields sentence from a source, that can be an iterable or a set of files.

    Warnings
    --------
    This class should not be used directly.
    Use derived classes instead.

    See Also
    --------
    :class:`TextGenerator`
    :class:`BrownGenerator`
    :class:`XmlGenerator`
    :class:`ConllGenerator`

    Parameters
    ----------
    source : a string or an iterable
        An iterable of sentences or a path to a file or a repository
    lower : boolean, optional
        If True (default), converts sentences to lower case
    digit : boolean, optional
        If True (default), replace numeric values with `DIGIT_TOKEN` in sentences
    ignore_punctuation: boolean, optional
        If True, the punctuation will be ignored when reading the corpus. Default : False
    """
    def __init__(self, source, lower=False, digit=False, ignore_punctuation=False):
        self.source = source
        self.annotated = False

        processors = []
        if lower:
            processors.append(self._lower)
        if digit:
            processors.append(self._digit)
        if ignore_punctuation:
            processors.append(self._remove_punctuation)

        self.transform = _compose(*processors)

    @abc.abstractstaticmethod
    def _lower(s):
        pass

    @abc.abstractstaticmethod
    def _digit(s):
        pass

    @abc.abstractstaticmethod
    def _remove_punctuation(s):
        pass

    @abc.abstractmethod
    def sentences(self):
        """Yields sentences from the source

        Yields
        -------
        list of str
        """
        pass


class TextSentenceGenerator(SentenceGenerator):
    """Sentence generator for simple text source

    See Also
    --------
    :class:`SentenceGenerator`
    """

    def sentences(self):
        with mangoes.utils.io.get_reader(self.source) as reader:
            for line in reader:
                try:
                    yield self.transform(line).split()
                except AttributeError:
                    yield [self.transform(token) for token in line]

    @staticmethod
    def _lower(s):
        return s.lower()

    @staticmethod
    def _digit(s):
        return re.sub(DIGIT_RE, DIGIT_TOKEN, s)

    @staticmethod
    def _remove_punctuation(s):
        return s.translate(PUNCTUATION_TRANSLATE_TABLE)


class AnnotatedSentenceGenerator(SentenceGenerator):
    """Base class for sentences generators from annotated source

    A sentence generator yields sentence from a source, that can be an iterable or a set of files.

    Warnings
    --------
    This class should not be used directly.
    Use derived classes instead.

    See Also
    --------
    :class:`SentenceGenerator`
    """
    FIELDS = ("form", "lemma", "POS")
    NUM_TAG = 'NUM'
    PUNCTUATION_TAG = 'PUNCT'

    class Token(namedtuple("Token", FIELDS)):
        def lower(self):
            return self.__class__(self.form.lower(), self.lemma, self.POS)

        def replace(self, value):
            return self.__class__(value, value, self.POS)

    Token.__new__.__defaults__ = (None,) * len(FIELDS)

    def __init__(self, source, lower=False, digit=False, ignore_punctuation=True):
        super().__init__(source, lower, digit, ignore_punctuation)
        self.annotated = True

    @abc.abstractmethod
    def sentences(self):
        pass

    @classmethod
    def _lower(cls, sentence):
        return [t.lower() for t in sentence]

    @classmethod
    def _digit(cls, sentence):
        return [t.replace(DIGIT_TOKEN) if t.POS == cls.NUM_TAG and re.match(DIGIT_RE, t.form) else t for t in sentence]

    @classmethod
    def _remove_punctuation(cls, sentence):
        return [t for t in sentence if t.POS != cls.PUNCTUATION_TAG]


class BrownSentenceGenerator(AnnotatedSentenceGenerator):
    """Sentence generator for text source annotated in Brown format

    See Also
    --------
    :class:`SentenceGenerator`
    """

    def sentences(self):
        with mangoes.utils.io.get_reader(self.source) as reader:
            for line in reader:
                try:
                    line = line.split()
                except AttributeError:
                    pass
                sentence = []
                for token in line:
                    word, pos, lemma = token.split('/')
                    sentence.append(self.Token(form=word, POS=pos, lemma=lemma))
                yield self.transform(sentence)


class XmlSentenceGenerator(AnnotatedSentenceGenerator):
    """Sentence generator for an XML source

    See Also
    --------
    :class:`SentenceGenerator`
    """
    FIELDS = ("id", "form", "lemma", "POS", "features", "head", "dependency_relation")

    class Token(namedtuple("Token", FIELDS)):
        def lower(self):
            return self.__class__(self.id, self.form.lower(), self.lemma, self.POS, self.features,
                                  self.head, self.dependency_relation)

        def replace(self, value):
            return self.__class__(self.id, value, value, self.POS, self.features,
                                  self.head, self.dependency_relation)

    Token.__new__.__defaults__ = ('_',) * len(FIELDS)

    def __init__(self, source, lower=False, digit=False, ignore_punctuation=False):
        super().__init__(source, lower, digit, ignore_punctuation)

        if os.path.exists(self.source):
            self.sentences = self._sentences_from_files
        else:
            self.sentences = self._sentences_from_string

    def _parse_token(self, xml_token, dependencies=None):
        token_id, governor_id, dep_type = '_', '_', '_'
        try:
            token_id = xml_token.attrib['id']
            try:
                for dep in dependencies.findall("dep"):  # dep[dependent/@idx={}]").format(token_id)
                    if dep.find('dependent').attrib['idx'] == token_id:
                        governor_id = dep.find('governor').attrib['idx']
                        dep_type = dep.attrib['type']
                        break
            except AttributeError:
                pass
        except KeyError:
            pass

        return self.Token(id=token_id,
                          form=xml_token.find('word').text,
                          POS=xml_token.find('POS').text,
                          lemma=xml_token.find('lemma').text,
                          head=governor_id,
                          dependency_relation=dep_type)

    def _sentences_from_xml(self, xml_sentences):
        for xml_sentence in xml_sentences:
            yield self.transform([self._parse_token(token,
                                                    xml_sentence.find("dependencies[@type='basic-dependencies']"))
                                  for token in xml_sentence.iter("token")])
        raise StopIteration

    def _sentences_from_files(self):
        for xml_file in mangoes.utils.io.recursive_list_files(self.source):
            tree = xml.etree.ElementTree.parse(xml_file)
            xml_sentences = tree.getroot().iter("sentence")
            yield from self._sentences_from_xml(xml_sentences)

    def _sentences_from_string(self):
        yield from self._sentences_from_xml(xml.etree.ElementTree.fromstring(self.source).iter("sentence"))


class ConllSentenceGenerator(AnnotatedSentenceGenerator):
    """Sentence generator for source annotated in Conll format

    See Also
    --------
    :class:`SentenceGenerator`
    """
    # Stanford Conll fields = ("id", "word", "lemma", "pos", "NER", "head", "deprel")
    # Mangoes fields =        ("id", "form", "lemma", "POS", "features", "head", "dependency_relation")
    FIELDS = ("id", "form", "lemma", "POS", "NER", "head", "dependency_relation")

    class Token(namedtuple("Token", FIELDS)):
        def lower(self):
            return self.__class__(self.id, self.form.lower(), self.lemma, self.POS, self.NER,
                                  self.head, self.dependency_relation)

        def replace(self, value):
            return self.__class__(self.id, value, value, self.POS, self.NER,
                                  self.head, self.dependency_relation)

    def _parse_token(self, line):
        return self.Token(*line.split('\t'))

    def sentences(self):
        with mangoes.utils.io.get_reader(self.source) as reader:
            sentence = []

            for line in reader:
                line = line.strip()
                if not line:
                    yield self.transform(sentence)
                    sentence = []
                elif line.startswith("#"):
                    pass
                else:
                    sentence.append(self._parse_token(line))
            if sentence:
                yield self.transform(sentence)


class ConllUSentenceGenerator(ConllSentenceGenerator):
    # Conll U fields = ("id", "form", "lemma", "upostag", "xpostag", "feats", "head", "deprel", "deps", "misc")
    # Mangoes fields = ("id", "form", "lemma", "POS", "features", "head", "dependency_relation")

    FIELDS = ("id", "form", "lemma", "POS", "xpostag", "feats", "head", "dependency_relation", "deps", "misc")

    class Token(namedtuple("Token", FIELDS)):
        def lower(self):
            return self.__class__(self.id, self.form.lower(), self.lemma, self.POS, self.xpostag, self.feats,
                                  self.head, self.dependency_relation, self.deps, self.misc)

        def replace(self, value):
            return self.__class__(self.id, value, value, self.POS, self.xpostag, self.feats,
                                  self.head, self.dependency_relation, self.deps, self.misc)



# -*- coding: utf-8 -*-
"""Functions to count the words co-occurrence within a corpus.

This module provides the main function count_cooccurrence to construct a CountBasedRepresentation.
"""
import collections
import logging
import math
import multiprocessing
import random

import numpy as np

import mangoes.context
import mangoes.utils.decorators
import mangoes.utils.exceptions
import mangoes.utils.multiproc
import mangoes.utils.counting
from mangoes.utils.options import ProgressBar

logger = logging.getLogger(__name__)


##########
# Computing / Building a cooccurrence count instance
##########
@mangoes.utils.decorators.timer(display=logger.info)
def count_cooccurrence(corpus, words,
                       context=mangoes.context.Window(),
                       subsampling=False, nb_workers=None, batch=1000000):
    """Build a CountBasedRepresentation where rows are the words in `words`, counting co-occurrences from the `corpus`.

    Examples
    --------
    >>> import mangoes.counting
    >>> window_5 = mangoes.context.Window(window_half_size=5)
    >>> counts_matrix = mangoes.counting.count_cooccurrence(corpus, vocabulary, context=window_5)

    Parameters
    -----------
    corpus: mangoes.Corpus
    words: mangoes.Vocabulary
        words represented as vectors (rows of the matrix)
    context: mangoes.context.Context or mangoes.Vocabulary
        A Vocabulary or context defining function such as defined in the :mod:`mangoes.context` module.
        Default is a window of size 1-x-1 : count the co-occurrences between the words in `words_vocabulary` and the
        words surrounding it.
        If `context` is a Vocabulary, only consider the words of this vocabulary in the window.
    nb_workers: int
        number of subprocess to use;
    subsampling: boolean or dict
        to apply subsampling on frequent words. Value can be False (default), True or a frequency
        threshold. If True, the default value of `create_subsampler()` function is used

    Returns
    -------
    mangoes.CountBasedRepresentation

    """
    if words is None:
        raise mangoes.utils.exceptions.RequiredValue("'words' is required to count cooccurrences")

    if isinstance(context, mangoes.Vocabulary):
        context = mangoes.context.Window(context)

    if not nb_workers:
        nb_workers = multiprocessing.cpu_count() - 1

    kwargs = {}
    if subsampling:
        kwargs['subsampler'] = create_subsampler(corpus) if subsampling is True else create_subsampler(corpus,
                                                                                                       subsampling)
    if corpus.nb_sentences:
        kwargs['nb_sentences'] = _estimate_nb_sentences_per_worker(corpus.nb_sentences, nb_workers)

    if nb_workers == 1:
        result = _count_context_cooccurrence(corpus.reader.sentences(),
                                             context, words, context.vocabulary, **kwargs).__next__()
        matrix, contexts_words = result
    else:
        data_parallel = mangoes.utils.multiproc.DataParallel(_count_context_cooccurrence,
                                                             mangoes.utils.counting._reduce_counter,
                                                             nb_workers=nb_workers, batch=batch)


        matrix, contexts_words = data_parallel.run(corpus,
                                                   *(context, words, context.vocabulary),
                                                   **kwargs)

    features = mangoes.Vocabulary(contexts_words)

    hyperparameters = {"corpus": corpus.params,
                       "context_definition": context.params,
                       "subsampling": subsampling}

    return mangoes.CountBasedRepresentation(words, features, matrix, hyperparameters)


def _count_context_cooccurrence(sentences, context, words_vocabulary, contexts_vocabulary,
                                nb_sentences=None, subsampler=None, batch=1000000):
    """Parallelizable function to count cooccurrence

    Count cooccurrence between words and contexts as defined by the 'context_definition' function, and on the
    sentences given by the 'sentences_enumerator'.

    Parameters
    -----------
    sentences: mangoes.utils.reader.SentenceGenerator
        an iterator over sentences
    context: callable
        a context defining function such as defined in 'mangoes.context'
    words_vocabulary: mangoes.Vocabulary
        words to represent as vectors (rows of the matrix)
    contexts_vocabulary: mangoes.Vocabulary
        words to use as features (columns of the matrix)
    nb_sentences: int, optional
        number of sentences. Only used for the progress bar
    subsampler: dict, optional
        a dictionary with probabilities of removal of frequent words

    Yields
    --------
    (collections.Counter, mangoes.Vocabulary)
    """
    counter = collections.Counter()

    build_contexts_vocabulary = contexts_vocabulary is None
    if build_contexts_vocabulary:
        contexts_vocabulary = mangoes.vocabulary.DynamicVocabulary()

    same_vocabulary = (words_vocabulary == contexts_vocabulary)

    filter_word_sentence = mangoes.vocabulary.create_tokens_filter(words_vocabulary.entity)
    filter_bigrams_sentence = mangoes.vocabulary.create_bigrams_filter(words_vocabulary.get_bigrams())

    context_accept_encoded_sentence = hasattr(context, 'call_on_encoded')
    if context_accept_encoded_sentence and not build_contexts_vocabulary:
        get_contexts = context.call_on_encoded
    else:
        get_contexts = context


    with ProgressBar(total=nb_sentences) as progress_bar:
        while True:
            try:
                for _ in range(batch):
                    sentence = sentences.__next__()
                    sentence = filter_bigrams_sentence(sentence)
                    if subsampler:
                        sentence = _subsample(sentence, subsampler)

                    word_sentence = filter_word_sentence(sentence)
                    word_sentence = words_vocabulary.indices(word_sentence)
                    word_sentence_mask = [True if w > -1 else False for w in word_sentence]

                    if build_contexts_vocabulary:
                        contexts_sentence = context.filter_sentence(sentence)
                        counter.update([(word, contexts_vocabulary.index(context_word))
                                        for word, word_contexts in zip(word_sentence, get_contexts(contexts_sentence))
                                        for context_word in word_contexts if word > -1])

                    else:
                        if same_vocabulary:
                            counter.update([(word, context_word)
                                            for word, word_contexts in
                                            zip(word_sentence, get_contexts(word_sentence, word_sentence_mask))
                                            for context_word in word_contexts if word > -1])
                        elif context_accept_encoded_sentence:
                            contexts_sentence = contexts_vocabulary.indices(context.filter_sentence(sentence))
                            counter.update([(word, context_word)
                                            for word, word_contexts in
                                            zip(word_sentence, get_contexts(contexts_sentence))
                                            for context_word in word_contexts if word > -1])
                        else:
                            contexts_sentence = sentence
                            counter.update([(word, contexts_vocabulary.index(context_word))
                                            for word, word_contexts in zip(word_sentence, get_contexts(contexts_sentence))
                                            for context_word in word_contexts if word > -1 ])

                    progress_bar.update()

                csr = mangoes.utils.counting._counter_to_csr(counter,
                                                             shape=(len(words_vocabulary), len(contexts_vocabulary)))
                yield (csr, contexts_vocabulary)

                counter = collections.Counter()
                if build_contexts_vocabulary:
                    contexts_vocabulary = mangoes.vocabulary.DynamicVocabulary()
            except StopIteration:
                break

    csr = mangoes.utils.counting._counter_to_csr(counter,
                                                 shape=(len(words_vocabulary), len(contexts_vocabulary)))
    yield (csr, contexts_vocabulary)


def _subsample(sentence, subsampler):
    return [word if word not in subsampler or random.random() > subsampler[word] else None for word in sentence]


def create_subsampler(corpus, threshold=10 ** -5):
    """Compute probabilities of removal of frequent words

    For each word appearing with a frequency higher than the threshold in the corpus, a probabilty of removal
    is computed following the formula :

    .. math::
            p = 1 - \sqrt{\\frac{t}{f}}

    where :math:`t` is the threshold and :math:`f` the frequency of the word in the corpus.

    Parameters
    ----------
    corpus: mangoes.Corpus
        Frequencies come from corpus.words_count
    threshold: float, optional
        Words appearing more than this threshold appear in the subsampler (default : :math:`10^{-5}`)

    Returns
    --------
    dict
        a dictionary associating each frequent word with a removal probability
    """
    threshold *= corpus.size
    return {word: 1 - math.sqrt(threshold / count) for word, count in corpus.words_count.items() if count > threshold}


def _estimate_nb_sentences_per_worker(nb_sentences, nb_workers):
    if nb_workers > 1:
        return nb_sentences // (nb_workers - 1)
    return nb_sentences


def merge(*counts, word_keys=None, context_keys=None, concat=lambda key, word: key + '_' + word):
    """Merge cooccurrence counts into one, providing parameters to handle how words and contexts should be merged

    Examples
    --------
    First example is a use case where your counts cooccurrences of words and POS tags from different languages
    >>> import mangoes
    >>> english_words = mangoes.Vocabulary(['can', 'car', 'cap'], language='en')
    >>> french_words = mangoes.Vocabulary(['car', 'cap'], language='fr')
    >>> pos_contexts = mangoes.Vocabulary(['ADJ', 'NOUN', 'VERB'])
    >>> en_count = mangoes.CountBasedRepresentation(english_words, pos_contexts, np.array(range(9)).reshape((3,3)))
    >>> print(en_count.to_df()) # doctest: +NORMALIZE_WHITESPACE
             ADJ  NOUN  VERB
        can    0     1     2
        car    3     4     5
        cap    6     7     8
    >>> fr_count = mangoes.CountBasedRepresentation(french_words, pos_contexts, np.array(range(6)).reshape((2,3)))
    >>> print(fr_count.to_df()) # doctest: +NORMALIZE_WHITESPACE
             ADJ  NOUN  VERB
        car    0     1     2
        cap    3     4     5
    >>> print(mangoes.counting.merge(en_count, fr_count, word_keys=True).to_df()) # doctest: +NORMALIZE_WHITESPACE
               ADJ  NOUN  VERB
        en_can   0     1     2
        en_car   3     4     5
        en_cap   6     7     8
        fr_car   0     1     2
        fr_cap   3     4     5

    A second example where contexts are keyed:
    >>> import mangoes
    >>> words = mangoes.Vocabulary(['a', 'b', 'c'])
    >>> contexts1 = mangoes.Vocabulary(['x', 'y', 'z'])
    >>> contexts2 = mangoes.Vocabulary(['x', 'y'])
    >>> count1 = mangoes.CountBasedRepresentation(words, contexts1, np.array(range(9)).reshape((3,3)))
    >>> print(count1.to_df()) # doctest: +NORMALIZE_WHITESPACE
             x     y     z
        a    0     1     2
        b    3     4     5
        c    6     7     8
    >>> count2 = mangoes.CountBasedRepresentation(words, contexts2, np.array(range(6)).reshape((3, 2)))
    >>> print(count2.to_df()) # doctest: +NORMALIZE_WHITESPACE
             x     y
        a    0     1
        b    2     3
        c    4     5
    >>> print(mangoes.counting.merge(count1, count2, context_keys=['c1', 'c2']).to_df()) # doctest: +NORMALIZE_WHITESPACE
            c1_x    c1_y    c1_z    c2_x    c2_y
        a      0       1       2       0       1
        b      3       4       5       2       3
        c      6       7       8       4       5
    >>> print(mangoes.counting.merge(count1, count2, context_keys=['c1', 'c2'], \
                                     concat=lambda k,w : '{}({})'.format(w, k)).to_df()) # doctest: +NORMALIZE_WHITESPACE
             x(c1)   y(c1)   z(c1)   x(c2)   y(c2)
        a      0       1       2       0       1
        b      3       4       5       2       3
        c      6       7       8       4       5

    And a third example with tokens
    >>> import collections
    >>> Token = collections.namedtuple('Token', 'lemma POS')
    >>> english_tokens = mangoes.Vocabulary([Token('can', 'N'), Token('can', 'V'), \
                                             Token('car', 'N'), Token('cap', 'N')], language='en')
    >>> french_tokens = mangoes.Vocabulary([Token('car', 'C'), Token('cap', 'N')], language='fr')
    >>> en_tok_count = mangoes.CountBasedRepresentation(english_tokens, pos_contexts, \
                                                        np.array(range(12)).reshape((4,3)))
    >>> print(en_tok_count.to_df()) # doctest: +NORMALIZE_WHITESPACE
                ADJ  NOUN  VERB
        can N    0     1     2
            V    3     4     5
        car N    6     7     8
        cap N    9    10    11
    >>> fr_tok_count = mangoes.CountBasedRepresentation(french_tokens, pos_contexts, np.array(range(6)).reshape((2,3)))
    >>> print(fr_tok_count.to_df()) # doctest: +NORMALIZE_WHITESPACE
                      ADJ  NOUN  VERB
        car C    0     1     2
        cap N    3     4     5
    >>> print(mangoes.counting.merge(en_tok_count, fr_tok_count, word_keys=True, \
                                     concat=lambda k, w: (*w, k)).to_df()) # doctest: +NORMALIZE_WHITESPACE
                   ADJ  NOUN  VERB
        can N en    0     1     2
            V en    3     4     5
        car N en    6     7     8
        cap N en    9    10    11
        car C fr    0     1     2
        cap N fr    3     4     5

    Parameters
    ----------
    counts: list of mangoes.CountBasedRepresentation
        List of cooccurrence counts to be merged
    word_keys: None (default), or bool or list of str
        If None or False, words that are common to several vocabularies are considered the same and their counts are
        summed.
        If word_keys is a list of string, of same size as counts, words are prefixed with these keys (prefixing is
        default but you can change that with format_str parameter).
        If word_keys is True, the languages of the vocabularies are used as keys
    context_keys: None (default), or bool or list of str
        If None or False, context words that are common to several context vocabularies are considered the same and
        their counts are summed.
        If context_keys is a list of string, of same size as counts, context words are prefixes with these keys.
        If context_keys is True, the languages of the context vocabularies are used as keys
    concat: callable, optional
        Function that takes a key and a word (or a token) as input and returns a new word (or token)
        If keys are given, this function is called to create the word of the merged vocabulary from the given keys and
        the original words
        Default is '{key}_{word}' that prefixes each word with their key and is only valid from simple string words
        vocabularies.
        Bigrams are transformed applying this function to both of their part

    Returns
    --------
    mangoes.CountBasedRepresentation
        a cooccurrence count with merged vocabulary, context and counts
    """
    mrg_vocabulary, vocabulary_mappings = mangoes.vocabulary.merge(*[c.words for c in counts], keys=word_keys,
                                                                   return_map=True, concat=concat)

    mrg_context, context_mappings = mangoes.vocabulary.merge(*[c.contexts_words for c in counts], keys=context_keys,
                                                             return_map=True, concat=concat)

    mrg_matrix = counts[0].matrix.copy()
    for count, row_mapping, context_mapping in zip(counts[1:], vocabulary_mappings[1:], context_mappings[1:]):
        mrg_matrix = mrg_matrix.combine(count.matrix, new_shape=(len(mrg_vocabulary), len(mrg_context)),
                                        row_indices_map=row_mapping, col_indices_map=context_mapping)

    return mangoes.CountBasedRepresentation(mrg_vocabulary, mrg_context, mrg_matrix)




# -*- coding: utf-8 -*-
"""Weighting transformations to apply to the co-occurrence count matrix.

This module provides transformations that can be used in the `transformations` parameter of
the :func:`mangoes.create_representation()` function to create an Embeddings from a CountBasedRepresentation.

Examples
---------

    >>> import mangoes.base
    >>> ppmi = mangoes.weighting.PPMI(alpha=1)
    >>> embeddings = mangoes.base.create_representation(cc, transformations=ppmi)


See Also
--------
:func:`mangoes.create_representation`
:class:`mangoes.Transformation`
"""

import numpy as np

import mangoes.utils.arrays
import mangoes.utils.exceptions
from mangoes.base import Transformation


class JointProbabilities(Transformation):
    """Defines a transformation that replaces counts with joint probabilities computed from these counts"""

    def __call__(self, matrix):
        """From matrix = [#(wi,cj)], compute the matrix [P(wi,cj)].

        Parameters
        ----------
        matrix: matrix-like object
            values must be non-negative integers

        Returns
        -------
        matrix-like object
        """
        mmatrix = mangoes.utils.arrays.Matrix.factory(matrix)
        return mmatrix / mmatrix.sum(axis=None)


class ConditionalProbabilities(Transformation):
    """Defines a transformation that replaces counts with joint probabilities computed from these counts"""

    def __call__(self, matrix):
        """From matrix = [#(wi,cj)], compute the matrix [P(cj|wi)].

        Assumes that 'matrix' contains only positive values, so that row sum normalization = row l1 normalization.

        Parameters
        ----------
        matrix: matrix-like object
            values must be non-negative integers

        Returns
        -------
        matrix-like object
        """
        return mangoes.utils.arrays.normalize(matrix, norm="l1", axis=1)


class ProbabilitiesRatio(Transformation):
    """Defines a transformation that replaces counts with joint probabilities computed from these counts

    Attributes
    ----------
    alpha: int, optional
        positive number (default=1); "smoothing" parameter for the computation of the context probability
        distributions: P(c_j) = (#c_j)**alpha / sum((#c_k)**alpha)
    """

    def __init__(self, alpha=1):
        super().__init__()
        self._params["alpha"] = alpha

    @property
    def alpha(self):
        return self._params["alpha"]

    def __call__(self, matrix):
        """From matrix = [#(wi,cj)], compute the matrix [P(cj|wi) / P(cj)].

        Parameters
        ----------
        matrix: matrix-like object
            values must be non-negative integers

        Returns
        -------
        matrix-like object
        """
        matrix = mangoes.utils.arrays.Matrix.factory(matrix)

        contexts_counts = matrix.sum(axis=0)
        if self.alpha != 1:
            contexts_counts = np.power(contexts_counts, self.alpha)

        nb_total_contexts = np.sum(contexts_counts)

        if nb_total_contexts > 0:
            contexts_proba = contexts_counts / nb_total_contexts
            contexts_proba[contexts_proba == 0] = 1  # For the upcoming division: this process is allowed
            # because in our case, if the inv_denominator is null then the numerator is theoretically null.
            probabilities = ConditionalProbabilities()
            contexts_words_dep_proba_matrix = probabilities(matrix)
            return contexts_words_dep_proba_matrix / contexts_proba
        else:
            return matrix


class PMI(Transformation):
    """Defines a PMI (Pointwise Mutual Information) transformation

    Attributes
    ----------
    alpha: int, optional
        positive number (default=1); "smoothing" parameter for the computation of the context probability
        distributions: P(c_j) = (#c_j)**alpha / sum((#c_k)**alpha)

    """

    def __init__(self, alpha=1):
        super().__init__()
        self._params["alpha"] = alpha

    @property
    def alpha(self):
        return self._params["alpha"]

    def __call__(self, matrix):
        """From matrix = [#(wi,cj)], compute the 'pmi' matrix [log(P(cj|wi)/P(cj)) or 0 if not defined].

        Parameters
        ----------
        matrix: matrix-like object
            values must be non-negative integers

        Returns
        -------
        matrix-like object
        """
        if self.alpha <= 0:
            raise mangoes.utils.exceptions.NotAllowedValue(msg="'alpha' parameter value must be positive "
                                                               "(value = {})".format(self.alpha))

        contexts_words_proba_ratio_matrix = ProbabilitiesRatio(alpha=self.alpha)(matrix)
        return np.log(contexts_words_proba_ratio_matrix.replace_negative_or_zeros_by(1))


class PPMI(PMI):
    """Defines a PPMI (Positive PMI) transformation.

    Attributes
    ----------
    alpha: int, optional
        positive number (default=1); "smoothing" parameter for the computation of the context probability
        distributions: P(c_j) = (#c_j)**alpha / sum((#c_k)**alpha)

    Returns
    -------
    matrix-like object
    """

    def __init__(self, alpha=1):
        super().__init__(alpha)

    def __call__(self, matrix):
        """From matrix = [#(wi,cj)], compute the 'ppmi' matrix [max(log(P(cj|wi)/P(cj)) or 0 if not defined, 0)].

        Parameters
        ----------
        matrix: matrix-like object
            values must be non-negative integers

        Returns
        -------
        matrix-like object
        """
        return ShiftedPPMI(alpha=self.alpha, shift=1)(matrix)


class ShiftedPPMI(PMI):
    """Defines a Shifted PPMI (Positive PMI) transformation

    From matrix = [#(wi,cj)], compute the matrix [max(log(P(cj|wi)/P(cj))-log(shift) or 0 if not defined, 0)]

    Attributes
    ----------
    alpha: int, optional
        positive number (default=1); "smoothing" parameter for the computation of the context probability
        distributions: P(c_j) = (#c_j)**alpha / sum((#c_k)**alpha)

    shift: int, optional
        number superior or equal to 1 (default=1); The result matrix will be shifted of log(shift)

    Returns
    -------
    matrix-like object
    """

    def __init__(self, alpha=1, shift=1):
        super().__init__(alpha)
        self._params["shift"] = shift

    @property
    def shift(self):
        return self._params["shift"]

    def __call__(self, matrix):
        """From matrix = [#(wi,cj)], compute the matrix [max(log(P(cj|wi)/P(cj))-log(shift) or 0 if not defined, 0)].

        Parameters
        ----------
        matrix: matrix-like object
            values must be non-negative integers

        Returns
        -------
        matrix-like object
        """
        if self.shift < 1:
            msg = "'shift' parameter value must be superior or equal to 1 (value = {})".format(self.shift)
            raise mangoes.utils.exceptions.NotAllowedValue(msg=msg)

        result = PMI(alpha=self.alpha)(matrix) - np.log(self.shift)
        return result.replace_negative_or_zeros_by(0)


class TFIDF(Transformation):
    """Defines a TF-IDF (term frequency-inverse document frequency) transformation"""

    def __call__(self, matrix):
        """From matrix = [#(wi,cj)], computes a tfidf matrix [tf(wi,cj) * idf(wi)]

        * tf(wi,cj) = P(cj|wi) or 0 if not defined
        * idf(wi) = log(#c / #(c,wj)) or 0 if not defined.

        Parameters
        ----------
        matrix: matrix-like object
            values must be non-negative integers

        Returns
        -------
        matrix-like object
        """
        mmatrix = mangoes.utils.arrays.Matrix.factory(matrix)

        tf_matrix = ConditionalProbabilities()(mmatrix)

        nb_contexts = mmatrix.shape[1]
        idf_array = mmatrix.nb_of_non_zeros_values_by_row()
        idf_array[(idf_array != 0)] = np.log(nb_contexts / idf_array[(idf_array != 0)])

        return tf_matrix.multiply_rowwise(idf_array)



# -*- coding: utf-8 -*-
"""
Class and functions to manage the documents used as corpus.
"""

import collections
import logging
import os.path
import pickle

import mangoes.utils
import mangoes.utils.decorators
import mangoes.utils.exceptions
import mangoes.utils.io
import mangoes.utils.reader
import mangoes.utils.counting

_logger = logging.getLogger(__name__)

TEXT = mangoes.utils.reader.TextSentenceGenerator
BROWN = mangoes.utils.reader.BrownSentenceGenerator
XML = mangoes.utils.reader.XmlSentenceGenerator
CONLL = mangoes.utils.reader.ConllSentenceGenerator
CONLLU = mangoes.utils.reader.ConllUSentenceGenerator


Token = collections.namedtuple('Token', ("form", "lemma", "POS"))


class Corpus:
    """Class to access to the source used as a Corpus

    The Corpus class creates a sentence generator from documents.


    Parameters
    ----------
    content : a string or an iterable
        An iterable of sentences or a path to a file or a repository
    name: str
        A name for your corpus.
        If no name is given and content is a path, the name will be this path.
    reader : class
        A class deriving from :class:`mangoes.utils.reader.SentenceGenerator`.
        Some shortcuts are defined in this module : TEXT (default), BROWN, XML and CONLL
    lower : boolean, optional
        If True, converts sentences to lower case. Default : False
    digit : boolean, optional
        If True, replace numeric values with the value of `DIGIT_TOKEN` in sentences. Default : False
    ignore_punctuation: boolean, optional
        If True, the punctuation will be ignored when reading the corpus. Default : False
    nb_sentences : int (optional)
        Expected number of sentences in Corpus, if known. This number is used to improve the output of the
        progress bar but the real value will be computed when initialized.
    lazy : boolean (optional)
        if False (default), count words and sentences when creating the Corpus (that can take a while);
        if True, only count words and sentences when needed.

    Attributes
    ----------
    words_count : collections.Counter
    nb_sentences
    size
    annotated : boolean
        Whether or not the corpus is annotated

    """

    def __init__(self, content, name=None, language=None, reader=TEXT,
                 lower=False, digit=False, ignore_punctuation=False,
                 nb_sentences=None, lazy=False):

        self._logger = logging.getLogger("{}.{}".format(__name__, self.__class__.__name__))

        self.content = content
        self.reader = reader(content, lower, digit, ignore_punctuation)
        self._reader_class = reader
        self.annotated = self.reader.annotated

        self._words_count = None
        self._bigrams_count = None
        self._size = None
        self._nb_sentences = nb_sentences

        if not name and isinstance(content, str) and os.path.exists(content):
            name = content

        self._params = {"lower": lower, "digit": digit, "ignore_punctuation": ignore_punctuation,
                        "name": name, "language": language}

        if not lazy:
            self._count_words()

    def __iter__(self):
        return self.reader.sentences()

    @property
    def nb_sentences(self):
        """Number of sentences in this corpus or None if unknown.

        Returns
        -------
        int or None

        Notes
        ------

        If Corpus is created with parameter lazy=True, this value is evaluated only when words_count is called.
        The value may also be set manually (if it is known).

        """
        return self._nb_sentences

    @nb_sentences.setter
    def nb_sentences(self, nb):
        self._nb_sentences = nb

    @property
    def name(self):
        """Name of the corpus"""
        return self._params["name"]

    @property
    def language(self):
        """Language of the corpus"""
        return self._params["language"]

    @property
    def lower(self):
        """If True, converts sentences to lower case."""
        return self._params["lower"]

    @property
    def digit(self):
        """If True, replace numeric values with the value of `DIGIT_TOKEN` in sentences."""
        return self._params["digit"]

    @property
    def ignore_punctuation(self):
        """If True, the punctuation will be ignored when reading the corpus."""
        return self._params["ignore_punctuation"]

    @property
    def params(self):
        """Parameters of the corpus"""
        return self._params

    @property
    def words_count(self):
        """Occurrences of each word in the corpus

        Returns
        -------
        collections.Counter
            a Counter with words as keys and number of occurrences as value

        """
        if self._words_count is None:
            self._count_words()
        return self._words_count

    @property
    def bigrams_count(self):
        """Occurrences of each bigram in the corpus

        Returns
        -------
        collections.Counter
            a Counter with bigrams as keys and number of occurrences as value

        """
        if self._bigrams_count is None:
            self._count_bigrams()
        return self._bigrams_count

    @property
    def size(self):
        """Number of words in the Corpus

        Returns
        -------
        int

        """
        if self._size is None:
            self._size = sum(self.words_count.values())

        return self._size

    def describe(self):
        """Print properties of this corpus"""
        print("Name:", self.name)
        if self.language:
            print("Language:", self.language)
        else:
            print("Language:", "unknown")
        print("Reader:", self._reader_class)
        print("Parameters:")
        for param in ["lower", "digit", "ignore_punctuation"]:
            print("\t-", param, ":", self.params[param])
        if self._size:
            print("Size:")
            print("\t- sentences:", self.nb_sentences)
            print("\t- total number of tokens:", self.size)
            print("\t- number of unique tokens:", len(self.words_count))
        else:
            print("The corpus has not been initialized yet.")

    def peek(self, size=5):
        """Print the first sentences of the corpus

        Parameters
        ----------
        size: int
            Number of sentences to display (default: 5)
        """
        sentences = self.reader.sentences()
        for _ in range(min(size, self.nb_sentences)):
            print(sentences.__next__())

    def create_vocabulary(self, attributes=None, filters=None):
        """Create a vocabulary from the corpus

        Parameters
        ----------
        attributes : string or tuple of string, optional
            If the Corpus is annotated, attribute(s) to get for each token. If None (default), all attributes are kept.
        filters : list of callables, optional
            A filter is a parametrized function that filter values from the Corpus' ``words_count``. This module
            provides 6 filters :
                * :func:`truncate`
                * :func:`remove_least_frequent`
                * :func:`remove_most_frequent`
                * :func:`remove_elements`
                * :func:`filter_by_attribute`
                * :func:`filter_attributes`

        Returns
        -------
        mangoes.Vocabulary
            The words are sorted by frequency

        Notes
        ------
        You can also write and use your own filters. A filter is a parametrized function that takes a
        ``collections.Counter()`` as input and returns a ``collections.Counter()``.
        It should be decorated with :func:`mangoes.utils.decorators.counter_filter`

        """

        if attributes:
            filter_token = mangoes.vocabulary.create_token_filter(attributes)
            words = collections.Counter()
            for token in self.words_count:
                words[filter_token(token)] += self.words_count[token]
        else:
            words = self.words_count

        if filters:
            for filter_ in filters:
                words = filter_(words)

        return mangoes.Vocabulary(sorted(words, key=words.get, reverse=True),
                                  entity=attributes, language=self.params["language"])

    def save_metadata(self, file_path):
        """Save metadata of this Corpus in a pickle file

        Save path to corpus, words_count, number of sentences, ... in a pickle file

        Parameters
        ----------
        file_path : string

        """
        with open(file_path, 'wb') as pickle_file:
            pickle.dump(self, pickle_file)

    @staticmethod
    def load_from_metadata(file_path):
        """Create a Corpus instance from previously saved metadata

        Parameters
        ----------
        file_path : string
        """
        with open(file_path, 'rb') as pickle_file:
            return pickle.load(pickle_file)

    @mangoes.utils.decorators.timer(display=_logger.info, label="counting occurrences of each word in corpus")
    def _count_words(self):
        if not self.annotated:
            self._words_count, self._nb_sentences = mangoes.utils.counting.count_words_raw(self, self._nb_sentences)
        else:
            self._words_count, self._nb_sentences = mangoes.utils.counting.count_words_annotated(self,
                                                                                                 self._nb_sentences)

    @mangoes.utils.decorators.timer(display=_logger.info, label="counting occurrences of each bigram in corpus")
    def _count_bigrams(self):
        if not self.annotated:
            self._bigrams_count, self._nb_sentences = mangoes.utils.counting.count_bigrams_raw(self, self._nb_sentences)
        else:
            self._bigrams_count, self._nb_sentences = mangoes.utils.counting.count_bigrams_annotated(self,
                                                                                                     self._nb_sentences)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_logger']
        del state['reader']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._logger = logging.getLogger("{}.{}".format(__name__, self.__class__.__name__))
        self.reader = self._reader_class(self.content, self.lower, self.digit, self.ignore_punctuation)


def _reduce_counter(total, part):
    total_counter, total_nb_sentences = total
    part_counter, part_nb_sentences = part

    total_counter.update(part_counter)
    total_nb_sentences += part_nb_sentences

    return total_counter, total_nb_sentences

# ##############################################
# Filters to select vocabulary from a corpus


@mangoes.utils.decorators.counter_filter
def truncate(max_nb, words_count):
    """Filter to apply to a counter to keep at most 'max_nb' elements.

    Elements with higher counts are preferred over elements with lesser counts.
    Elements with equal counts are arbitrarily selected during the truncating, if necessary.

    Parameters
    ----------
    max_nb: positive int
        Maximal number of elements to keep
    words_count: collections.Counter
        The counter to filter

    Returns
    --------
    collections.Counter

    See Also
    ----------
    :class:`mangoes.vocabulary.Vocabulary`
    :func:`mangoes.utils.decorators.counter_filter`


    """
    return collections.Counter({word: frequency
                                for (word, frequency) in words_count.most_common(max_nb)})


@mangoes.utils.decorators.counter_filter
def remove_least_frequent(min_frequency, words_count):
    """Filter to apply to a counter to keep the elements with a high enough frequency.

    Parameters
    ----------
    min_frequency: positive int or float
        If >= 1, will be interpreted as a 'count' value (a positive integer), else,
        will be interpreted as a frequency.
    words_count: collections.Counter
        The counter to filter

    Returns
    --------
    collections.Counter

    Examples
    ---------
    >>> vocabulary = mangoes.Vocabulary(corpus,
    >>>                                filters=[mangoes.vocabulary.remove_least_frequent(min_frequency)])

    See Also
    ----------
    :class:`mangoes.vocabulary.Vocabulary`
    :func:`mangoes.utils.decorators.counter_filter`
    """

    if min_frequency >= 1:
        return collections.Counter({w: freq
                                    for w, freq in words_count.items() if freq >= min_frequency})
    if min_frequency >= 0:
        total_number = float(sum(words_count.values()))
        return collections.Counter({w: freq
                                    for w, freq in words_count.items() if
                                    freq / total_number >= min_frequency})

    raise mangoes.utils.exceptions.NotAllowedValue(msg="'min_frequency' parameter value must be positive "
                                                       "(value = {}".format(min_frequency))


@mangoes.utils.decorators.counter_filter
def remove_most_frequent(max_frequency, words_count):
    """Filter to apply to a counter to only keep the elements with a low enough frequency.

    Parameters
    ----------
    max_frequency: positive int or float
        If >= 1, will be interpreted as a 'count' value (a positive integer), else,
        will be interpreted as a frequency.
    words_count: collections.Counter
        The counter to filter

    Returns
    --------
    collections.Counter

    Examples
    ---------
    >>> vocabulary = mangoes.Vocabulary(corpus,
    >>>                                filters=[mangoes.vocabulary.remove_most_frequent(max_frequency)])

    See Also
    ----------
    :class:`mangoes.vocabulary.Vocabulary`
    :func:`mangoes.utils.decorators.counter_filter`
    """
    if max_frequency >= 1:
        return collections.Counter(
            {w: freq for w, freq in words_count.items() if freq <= max_frequency})
    elif max_frequency >= 0:
        total_number = float(sum(words_count.values()))
        return collections.Counter(
            {w: freq for w, freq in words_count.items() if freq / total_number <= max_frequency})

    raise mangoes.utils.exceptions.NotAllowedValue(msg="'max_frequency' parameter value must be positive "
                                                       "(value = {}".format(max_frequency))


@mangoes.utils.decorators.counter_filter
def remove_elements(stopwords, words_count=None, attribute=None):
    """Filter to apply to a counter to remove the elements in 'stopwords' set-like object.

    Parameters
    -----------
    stopwords: list or set or string
        collection of words to remove from the words_count
        (ex: nltk.corpus.stopwords.words("english") or string.punctuation)
    attribute: str or tuple, optional
        If the keys in words_count are annotated tokens, attribute to consider
    words_count: collections.Counter
        The counter to filter

    Returns
    --------
    collections.Counter

    See Also
    ----------
    :class:`mangoes.vocabulary.Vocabulary`
    :func:`mangoes.utils.decorators.counter_filter`
    """
    if attribute:
        return collections.Counter({w: freq for w, freq in words_count.items()
                                    if getattr(w, attribute) not in stopwords})
    return collections.Counter({w: freq for w, freq in words_count.items() if w not in stopwords})


@mangoes.utils.decorators.counter_filter
def filter_by_attribute(attribute, value, words_count=None):
    """Filter to apply to a counter to only keep certain tokens, based on the value of an attribute.

    This filter can only be applied to an annotated Corpus

    Parameters
    -----------
    attribute: str
        If the keys in words_count are annotated tokens, attribute to consider
    value: string or set of strings
        List of the values to keep for the attribute
    words_count: collections.Counter
        The counter to filter

    Returns
    --------
    collections.Counter

    See Also
    ----------
    :class:`mangoes.vocabulary.Vocabulary`
    :func:`mangoes.utils.decorators.counter_filter`
    """
    if isinstance(value, str):
        value = [value]
    return collections.Counter({w: freq for w, freq in words_count.items()
                                if getattr(w, attribute) in value})


@mangoes.utils.decorators.counter_filter
def filter_attributes(attributes, words_count=None):
    """Filter to apply to a counter to only keep certain attributes of the tokens.

    This filter can only be applied to an annotated Corpus

    Examples
    --------
    >>> import collections
    >>> Token = collections.namedtuple('Token', ('form', 'lemma', 'POS'))
    >>> words_count = {Token('can', 'can', 'NOUN'): 5, Token('cans', 'can', 'NOUN'): 2, Token('can', 'can', 'VBZ'): 3}
    >>> filter_attributes('lemma', words_count)
    Counter({'can': 10})
    >>> filter_attributes(('lemma', 'POS'), words_count)
    Counter({Token(lemma='can', POS='NOUN'): 7, Token(lemma='can', POS='VBZ'): 3})

    Parameters
    -----------
    attributes: str or tuple of str
        If the keys in words_count are annotated tokens, attributes to keep
    words_count: collections.Counter
        The counter to filter

    Returns
    --------
    collections.Counter

    See Also
    ----------
    :class:`mangoes.vocabulary.Vocabulary`
    :func:`mangoes.utils.decorators.counter_filter`
    """
    token_filter = mangoes.vocabulary.create_token_filter(attributes)
    new_words_count = collections.Counter()
    for w in words_count:
        new_words_count[token_filter(w)] += words_count[w]
    return new_words_count



# -*- coding: utf-8 -*-
"""Utility classes and functions to handle matrices

This module provides the :class:`Matrix` class that encapsulates all needed methods for both sparse and dense matrices.

"""
import abc
import logging
import os
import pickle
import warnings

import numpy as np
from numpy import dtype
import sklearn.preprocessing
from sklearn.utils import DataConversionWarning
from scipy import sparse

import mangoes.utils.exceptions

logger = logging.getLogger(__name__)
warnings.simplefilter('ignore', DataConversionWarning)


# Because we will normalize and / or center matrices with an 'int' dtype


def sqrt(matrix, inplace=False):
    """Perform element-wise sqrt.

    Parameters
    ----------
    matrix: a matrix-like object
    inplace: boolean, optional
        whether or not to perform the operation inplace (default=False)

    Returns
    -------
    a matrix-like object
        has the same type as the one input
    """
    return matrix.sqrt(inplace=inplace)


def normalize(matrix, norm="l2", axis=1, inplace=False):
    """Normalize the matrix-like input.

    Parameters
    ----------
    matrix: a matrix-like object
    norm: {'l2' (default), 'l1' or 'max'}, optional
    axis: {1 (default) or 0}
        specify along which axis to compute the average values that will be used to carry out the normalization.
        Use axis=1 to normalize the rows, and axis=0 to normalize the columns.
    inplace: boolean
        whether or not to perform inplace normalization (default=False). Note that if the sparse matrix's format is
        not compatible with the axis along which the normalization is asked (csr for axis=1, csc for axis=0), another
        one will be created, and 'inplace' will become moot. Also, will not apply if the input's dtype is 'int' instead
        of a type compatible with division, such as 'float'.

    Returns
    -------
    a matrix-like object

    Warnings
    --------
    the inplace=True parameter isn't implemented yet

    """
    return matrix.normalize(norm=norm, axis=axis, inplace=inplace)


def center(matrix, axis=1, inplace=False):
    """Center the matrix-like input wrt row average, column average, or total average.

    Parameters
    ----------
    matrix: a matrix-like object
    axis: {1 (default), 0, None}
        the axis along which to compute the average vector (ex: if axis=0, the columns are centered, if axis=1,
        then the rows are). If None, the average is computed using all the values of the matrix, and is then subtracted
        from every value.
    inplace: boolean
        whether or not to perform inplace normalization (default=False). Note that if the sparse matrix's format is not
        compatible with the axis along which the normalization is asked (csr for rows=True, csc for Rows=False), another
        one will be created, and 'inplace' will become moot. Also, will not apply if the input's dtype is 'int' instead
        of a type compatible with division, such as 'float'.

    Returns
    -------
    a matrix-like object

    Warnings
    --------

    * the inplace=True parameter isn't implemented yet
    * If the input is a sparse matrix, then a dense numpy array will be created (and the 'inplace' option becomes
      moot), possibly causing a memory exhaustion error.
    """
    return matrix.center(axis=axis, inplace=inplace)


class Matrix:
    """Abstract class used to store generated vectors in a matrix

    """

    @abc.abstractmethod
    def multiply_rowwise(self, array):
        """Multiply the values of the matrix by the value of the array whose index corresponds to the row index

        Examples
        --------
        >>> M = mangoes.utils.arrays.Matrix.factory(np.asarray([[1, 1, 1], [2, 2, 2]]))
        >>> v = [[3], [4]]
        >>> M.multiply_rowwise(v)
        NumpyMatrix([[3, 3, 3], [8, 8, 8]])

        Parameters
        ----------
        array: array
            a 1d array with length = nb of matrix rows

        Returns
        -------
        matrix
            new matrix with same type as self
        """
        pass

    @abc.abstractmethod
    def normalize(self, norm="l2", axis=1, inplace=False):
        """Normalize the matrix.

        Parameters
        ----------
        norm: {'l2' (default), 'l1' or 'max'}
        axis: {1 (default) or 0}
            specify along which axis to compute the average values that will be used to carry out the normalization.
            Use axis=1 to normalize the rows, and axis=0 to normalize the columns.
        inplace: boolean
            whether or not to perform inplace normalization (default=False). Note that if the sparse matrix's format is
            not compatible with the axis along which the normalization is asked (csr for axis=1, csc for axis=0),
            another one will be created, and 'inplace' will become moot. Also, will not apply if the input's dtype
            is 'int' instead of a type compatible with division, such as 'float'.

        Returns
        -------
        a matrix-like object

        Warnings
        --------
        the inplace=True parameter isn't implemented yet

        """
        pass

    @abc.abstractmethod
    def center(self, axis=1, inplace=False):
        """Center the matrix to row average, column average, or total average.

        Parameters
        ----------
        axis: {1 (default), None, 0}
            the axis along which to compute the average vector (ex: if axis=0, the columns  are centered, if axis=1,
            then the rows are). If None, the average is computed using all the values of the matrix, and is then
            subtracted from every value.
        inplace: boolean
            whether or not to perform inplace normalization (default=False). Note that if the sparse matrix's format is
            not compatible with the axis along which the normalization is asked (csr for rows=True, csc for Rows=False),
            another one will be created, and 'inplace' will become moot. Also, will not apply if the input's dtype
            is 'int' instead of a type compatible with division, such as 'float'.

        Returns
        -------
        a matrix-like object

        Warnings
        --------
        * the inplace=True parameter isn't implemented yet
        * If the input is a sparse matrix, then a dense numpy array will be created (and the 'inplace' option becomes
        moot), possibly causing a memory exhaustion error.
        """
        pass

    @abc.abstractmethod
    def sqrt(self, inplace=False):
        """Performs element-wise sqrt.

        Parameters
        ----------
        inplace: boolean
            whether or not to perform the operation inplace (default=False)

        Returns
        -------
        a matrix-like object
        """
        pass

    @abc.abstractmethod
    def combine(self, other, new_shape, row_indices_map=None, col_indices_map=None):
        """Merge another matrix with this one, creating a new matrix

        The shape of the merged matrix is `new_shape`
        The rows and columns of the other matrix are mapped to the columns of the merged matrix using `xxx_indices_map`
        and the resulting matrix is added to the current one, extended to fit the new shape.

        Examples
        --------
        >>> A = NumpyMatrix(np.array(range(6)).reshape((2,3)))
        >>> print(A)
        [[0 1 2]
         [3 4 5]]
        >>> B = NumpyMatrix(np.array(range(4)).reshape((2,2)))
        >>> print(B)
        [[0 1]
         [2 3]]
        >>> A.combine(B, (2, 4), col_indices_map={0:1, 1:3})
        NumpyMatrix([[ 0.,  1.,  2.,  1.],
                     [ 3.,  6.,  5.,  3.]])

        indices_map = {0:1,     # column 0 of the second matrix go to column 1 of the merged matrix
                       1:3}     # column 1 of the second matrix go to column 3 of the merged matrix

        Parameters
        ----------
        other: matrix-like object
            another matrix to merge with this one
        new_shape: tuple
            the shape of the resulting matrix
        row_indices_map: dict
            a mapping between the indices of the rows of 'other' and the merged matrix
        col_indices_map: dict
            a mapping between the indices of the columns of 'other' and the merged matrix

        Returns
        -------
        merged : matrix-like object

        """
        pass

    @abc.abstractmethod
    def pairwise_distances(self, Y=None, metric="cosine", **kwargs):
        """ Compute a distance matrix from the rows of the matrix.

        This method returns the matrix of the distances between the rows of the matrix.
        Or, if `Y` is given (default is None), then the returned matrix is the pairwise distance between the rows of
        the matrix and the ones from Y.

        This function relies on the `sklearn.metrics.pairwise_distances` module so you can use any distance
        available in it.

        Valid values for metric are:
        - From scikit-learn: ['cityblock', 'cosine', 'euclidean', 'l1', 'l2',
          'manhattan']. These metrics support sparse matrix inputs.
        - From scipy.spatial.distance: ['braycurtis', 'canberra', 'chebyshev',
          'correlation', 'dice', 'hamming', 'jaccard', 'kulsinski', 'mahalanobis',
          'matching', 'minkowski', 'rogerstanimoto', 'russellrao', 'seuclidean',
          'sokalmichener', 'sokalsneath', 'sqeuclidean', 'yule']

        See the documentation for sklearn.metrics.pairwise_distances for details on these metrics.

        Parameters
        ----------
        Y: matrix, optional
            An optional second matrix.
        metric : string, or callable
            The metric to use when calculating distance between vectors.
        **kwds : optional keyword parameters
            Any further parameters are passed directly to the distance function.

        Returns
        -------
        array [len(self), len(self)] or [len(self), len(Y)]
            A distance matrix D such that D_{i, j} is the distance between the ith and the jth rows of this matrix,
            if Y is None.
            If Y is not None, then D_{i, j} is the distance between the ith row of this matrix and the jth row of Y.
        """
        pass

    @abc.abstractmethod
    def replace_negative_or_zeros_by(self, value):
        """Replace the negative or zeros value of the matrix by a new value

        Parameters
        ----------
        value: int or float

        Returns
        -------
        a matrix-like object
        """
        pass

    @abc.abstractmethod
    def nb_of_non_zeros_values_by_row(self):
        """Return the numbers of non zeros value for each line of the matrix

        Returns
        -------
        np.array
            np.array with number of non zeros values for each row
        """
        pass

    @abc.abstractmethod
    def all_positive(self):
        """Check if all values of the matrix are positive or zero.

        Returns
        -------
        boolean
        """

    @abc.abstractmethod
    def as_dense(self):
        """Return a dense version of the matrix

        Returns
        -------
        array
        """

    @abc.abstractmethod
    def save(self, path, name="matrix"):
        """Save the matrix in a file.

        The format of the file depends of the type of the matrix. See subclasses for more information.

        Parameters
        ----------
        path: str
            path to the folder where the file will be saved
        name: str
            name of the file (without extension)

        Returns
        -------
        str
            Complete path to the created file

        """
        pass

    @classmethod
    def load(cls, path, name):
        """Load a matrix from a '.npz' archive of '.npy' files.

        Parameters
        ----------
        path: str
            path to the folder where the archive is stored
        name: str
            name of the file (without extension)

        Returns
        -------
        a matrix-like object

        """
        try:
            return csrSparseMatrix.load(path, name)
        except FileNotFoundError:
            return NumpyMatrix.load(path, name)

    @abc.abstractstaticmethod
    def format_vector(vector, sep):
        """Format a line of a csrSparseMatrix as a string

        Parameters
        ----------
        vector:
            line of a csrSparseMatrix
        sep: str
            separator to use

        Returns
        -------
        str
            string representation of the vector
        """
        pass

    @staticmethod
    def factory(matrix):
        """Create a Matrix object from a numpy.ndarray or a scipy.sparse.csr_matrix

        If matrix is already a Matrix, returns it

        Parameters
        ----------
        matrix:
            Matrix, numpy.ndarray or scipy.sparse.csr_matrix

        Returns
        -------
        a matrix-like object
        """
        if isinstance(matrix, Matrix):
            return matrix
        if isinstance(matrix, np.ndarray):
            return NumpyMatrix(matrix)
        if isinstance(matrix, sparse.csr_matrix):
            return csrSparseMatrix(matrix)
        if isinstance(matrix, int) or isinstance(matrix, float):
            return matrix

        raise mangoes.utils.exceptions.UnsupportedType("Can't create a matrix from {}".format(type(matrix)))


class csrSparseMatrix(sparse.csr_matrix, Matrix):
    # use of an invalid name because scipy uses the first 3 letters of the class name
    """Class used to store generated vectors in a matrix with a scipy.sparse.csr_matrix

    """
    EXT = '.npz'

    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)
        self.is_sparse = True

    def multiply_rowwise(self, array):
        nonzero_rows, _ = self.nonzero()
        result = self._with_data(self.data * array[nonzero_rows], copy=True)
        result.eliminate_zeros()
        return csrSparseMatrix(result)

    def __truediv__(self, other):
        import scipy
        if scipy.isscalar(other):
            result = sparse.csr_matrix(self, dtype=np.float64)
            result.data /= other
            result.eliminate_zeros()
            return csrSparseMatrix(result)
        inv_denominator = sparse.csr_matrix(1.0 / other)
        result = self.multiply(inv_denominator)
        result.eliminate_zeros()
        return csrSparseMatrix(result)

    def __sub__(self, other):
        if np.isscalar(other):
            result = self._with_data(self.data - other, copy=True)
            result.eliminate_zeros()
            return result
        return super(csrSparseMatrix, self).__sub__(other)

    def __getitem__(self, item):
        return Matrix.factory(super(csrSparseMatrix, self).__getitem__(item))

    def normalize(self, norm="l2", axis=1, inplace=False):
        return sklearn.preprocessing.normalize(self, norm=norm, axis=axis, copy=not inplace)

    def center(self, axis=1, inplace=False):
        if inplace:
            raise NotImplementedError  # TODO : inplace

        matrix = self.A
        if axis is None:
            return matrix - np.mean(matrix, axis=None)
        else:
            return Matrix.factory(sklearn.preprocessing.scale(matrix, axis=axis, with_mean=True,
                                                              with_std=False, copy=False))

    def sqrt(self, inplace=False):
        if inplace:
            self.data = np.sqrt(self.data)
            return self
        else:
            return self._with_data(np.sqrt(self.data), copy=not inplace)

    def log(self):
        """Natural logarithm, element-wise.

        Adaptation of numpy.log to csr_matrix
        """
        return self._with_data(np.log(self.data), copy=True)

    def replace_negative_or_zeros_by(self, value):
        data = self.data
        data[data <= 0] = value
        result = self._with_data(data, copy=True)
        result.eliminate_zeros()
        return result

    def nb_of_non_zeros_values_by_row(self):
        return np.array(np.diff(self.indptr), dtype=np.float)

    def all_positive(self):
        return np.all(self.data >= 0)

    def combine(self, other, new_shape, row_indices_map=None, col_indices_map=None):
        if not row_indices_map and not col_indices_map:
            return Matrix.factory(self + other)

        get_row = (lambda r: row_indices_map[r]) if row_indices_map else (lambda i: i)
        get_col = (lambda c: col_indices_map[c]) if col_indices_map else (lambda j: j)

        mrg_matrix = self.tolil()
        mrg_matrix.rows = np.resize(mrg_matrix.rows, new_shape[0])
        mrg_matrix.data = np.resize(mrg_matrix.data, new_shape[0])
        for i in range(mrg_matrix.shape[0], new_shape[0]):
            mrg_matrix.rows[i] = []
            mrg_matrix.data[i] = []
        mrg_matrix._shape = new_shape

        for (i, j), v in other.todok().items():
            mrg_matrix[get_row(i), get_col(j)] += v

        return Matrix.factory(mrg_matrix.tocsr())

    def as_dense(self):
        return self.A

    def save(self, path, name='matrix'):
        """Save the matrix as a '.npz' archive of '.npy' files.

        Parameters
        ----------
        path: str
            path to the folder where the archive will be written
        name: str
            name of the file (without extension)

        Returns
        -------
        str
            Complete path to the created file

        """
        file_path = os.path.join(path, name + '.npz')
        np.savez_compressed(file_path, data=self.data, indices=self.indices,
                            indptr=self.indptr, shape=self.shape)
        return file_path

    @classmethod
    def load(cls, path, name):
        """Load a _matrix from a '.npz' archive of '.npy' files.

        Parameters
        ----------
        path: str
            path to the folder where the archive is stored
        name: str
            name of the file (without extension)

        Returns
        -------
        csrSparseMatrix

        """
        loader = np.load(os.path.join(path, name + cls.EXT))
        return cls((loader['data'], loader['indices'], loader['indptr']),
                   shape=loader['shape'])

    @staticmethod
    def load_from_text_file(file_object, nb_columns, data_type, sep):
        """Load a matrix and a list of words from a text file

        Parameters
        ----------
        file_object: file-like object
        nb_columns: int
            number of columns in the matrix
        data_type
        sep: str
            token used as column separator in the text file

        Returns
        -------
        csrSparseMatrix
        """
        # TODO : move to Embeddings
        words = []
        data_list = []
        indices_list = []
        indptr = [0]

        for line in file_object:
            word, *string_numbers = line.strip().split(sep=sep)
            indices, data = zip(*(s.split(":") for s in string_numbers))

            words.append(word)
            data_list.extend(data)
            indices_list.extend(np.array(indices, dtype=dtype('int64')))
            indptr.append(indptr[-1] + len(indices))

        matrix = sparse.csr_matrix((data_list, indices_list, indptr),
                                   shape=((len(words)), nb_columns), dtype=data_type)
        return matrix, words

    @staticmethod
    def format_vector(vector, sep):
        return sep.join("{:d}:{:s}".format(i, str(vector[0, i])) for i in vector.nonzero()[1])

    @staticmethod
    def allclose(first, second):
        """Returns True if two arrays are element-wise equal within a tolerance.git

        Parameters
        ----------
        first: matrix
        second: matrix

        Returns
        -------
        bool
        """
        if not sparse.issparse(first) or not sparse.issparse(second):
            raise mangoes.utils.exceptions.IncompatibleValue("One of the objects is not a scipy.sparse matrix.")
        if first.shape != second.shape:
            return False
        diff_matrix = first - second
        return np.allclose(np.sum(np.abs(diff_matrix.data), None), 0)

    def pairwise_distances(self, Y=None, metric="cosine", **kwargs):
        if Y is not None:
            Y = _check_array(Y, ensure_2d=True)
        if metric == 'chebyshev':
            return sklearn.metrics.pairwise_distances(self, Y, metric=self.chebyshev)

        return sklearn.metrics.pairwise_distances(self, Y, metric=metric, **kwargs)

    @staticmethod
    def chebyshev(X, Y):
        diff = X - Y
        diff.data = np.abs(diff.data)
        return diff.max(axis=1).A


class NumpyMatrix(np.ndarray, Matrix):
    """Class used to store generated vectors in a matrix with a numpy.ndarray

    """
    EXT = '.npy'

    # subclassing np.ndarray : https://docs.scipy.org/doc/np/user/basics.subclassing.html
    def __new__(cls, input_array, info=None):
        obj = np.asarray(input_array).view(cls)
        obj.is_sparse = False
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.is_sparse = getattr(obj, 'is_sparse', None)

    def multiply_rowwise(self, array):
        return self * array

    def normalize(self, norm="l2", axis=1, inplace=False):
        if inplace:
            raise NotImplementedError  # TODO : inplace
        if len(self.shape) == 1:
            # sklearn's normalize function does not work in that case
            matrix = self._check_np_array(self, copy=not inplace,
                                          dtype=(np.float64, np.float32, np.float16))

            if norm == "l2":
                denominator = np.sqrt(np.sum(np.power(matrix, 2)))
            elif norm == "l1":
                denominator = np.sum(np.abs(matrix))
            elif norm == "max":
                denominator = np.max(np.abs(matrix))
            else:
                raise mangoes.utils.exceptions.NotAllowedValue(value=norm, allowed_values=["l1", "l2", "max"])
            if denominator > 0:
                matrix /= denominator
            return matrix
        else:
            return self.__class__(sklearn.preprocessing.normalize(self, norm=norm, axis=axis, copy=not inplace))

    def center(self, axis=1, inplace=False):
        if inplace:
            raise NotImplementedError  # TODO : inplace

        if axis is None:
            return self - np.mean(self, axis=None)
        else:
            return self.__class__(
                sklearn.preprocessing.scale(self, axis=axis, with_mean=True, with_std=False, copy=False))

    def sqrt(self, inplace=False):
        if not inplace:
            return np.sqrt(self)
        else:
            np.sqrt(self, self)
            return self

    def replace_negative_or_zeros_by(self, value):
        result = self.copy()
        result[(result <= 0)] = value
        return result

    def nb_of_non_zeros_values_by_row(self):
        return np.array([(self != 0).sum(axis=1)], dtype=np.float64).T

    def all_positive(self):
        return np.all(self >= 0)

    def combine(self, other, new_shape, row_indices_map=None, col_indices_map=None):
        if not row_indices_map and not col_indices_map:
            merged_matrix = self + other
        else:
            merged_matrix = np.zeros(new_shape)
            if col_indices_map and not row_indices_map:
                for old_index, new_index in col_indices_map.items():
                    merged_matrix[:, new_index] = other[:, old_index]
                merged_matrix[:, :self.shape[1]] += self
            elif row_indices_map and not col_indices_map:
                for old_index, new_index in row_indices_map.items():
                    merged_matrix[new_index, :] = other[old_index, :]
                merged_matrix[:self.shape[0], :] += self
            else:
                for (i, j), value in np.ndenumerate(other):
                    merged_matrix[row_indices_map[i], col_indices_map[j]] = value
                merged_matrix[:self.shape[0], :self.shape[1]] += self

        return Matrix.factory(merged_matrix)

    def as_dense(self):
        return self

    def save(self, path, name='matrix'):
        file_path = os.path.join(path, name + '.npy')
        np.save(file_path, self)
        return file_path

    @staticmethod
    def format_vector(vector, sep):
        return sep.join(str(v) for v in vector)

    @staticmethod
    def load_from_text_file(file_object, sep):
        """Load a matrix and a list of words from a text file

        Parameters
        ----------
        file_object: file-like object
        sep: str
            token used as a separator in the text file

        Returns
        -------
        tuple
            (matrix, list of words)
        """
        # TODO : move to Embeddings
        words = []
        vectors = []

        for line in file_object:
            word, *string_numbers = line.strip().split(sep=sep)
            words.append(word)
            vectors.append(np.array(string_numbers, dtype=np.float))

        matrix = np.array(vectors, dtype=np.float)
        return matrix, words

    @classmethod
    def load(cls, path, name):
        try:
            return np.asarray(np.load(os.path.join(path, name + cls.EXT)))
        except OSError as error:
            # If we have the following error, then the file is probably the result of
            # np.ndarray.dump, which is just a wrapper around a pickle file
            if str(error) == "Failed to interpret file '{}' as a pickle".format(path):
                with open(path, "rb") as dump_file:
                    matrix = np.asarray(pickle.load(dump_file, encoding="latin1"))
                warning_message = """Problem in 'np.load({:s})': in order to try to load
                                  content, the file will now be interpreted as a latin1 encoded
                                  pickle file.
                                  \n\t\t\t\t\tThis can happen if the file is the result of
                                  'np.ndarray.dump()' in python2. To prevent such
                                  compatibility issues,  prefer using 'np.save()' rather
                                  than 'np.ndarray.dump()' if possible."""
                logging.warning(warning_message, path)
                return matrix

    @staticmethod
    def _check_np_array(array, copy=False, dtype="numeric", order=None):
        """Check whether or not a np array can be correctly divided inplace.

        Comes from sklearn's 'check_array' function, but without the constraint of being limited to 2D
        arrays, which will soon becomes cause for exception raising according to sklearn's code.

        Parameters
        ----------
        array: numpy.array
            np array to check / convert
        copy: bool
            whether a forced copy will be triggered. If copy=False (default), a copy might still be triggered by a
            conversion.
        dtype:  str, type, list of types or None
            data type of result (default="numeric").
            If None, the dtype of the input is preserved. If "numeric", dtype is preserved unless array.dtype is object.
            If dtype is a list of types, conversion on the first type is only performed if the dtype of the input is
            not in the list.
        order:  {None (default), 'F', 'C'}
            whether an array will be forced to be fortran or c-style.

        Returns
        -------
        np.ndarray
            a 'correct' version of the input array
        """
        dtype_numeric = dtype == "numeric"

        dtype_orig = getattr(array, "dtype", None)
        if not hasattr(dtype_orig, 'kind'):
            # not a data type (e.g. a column named dtype in a pandas DataFrame)
            dtype_orig = None

        if dtype_numeric:
            if dtype_orig is not None and dtype_orig.kind == "O":
                # if input is object, convert to float.
                dtype = np.float64
            else:
                dtype = None

        if isinstance(dtype, (list, tuple)):
            if dtype_orig is not None and dtype_orig in dtype:
                # no dtype conversion required
                dtype = None
            else:
                # dtype conversion required. Let's select the first element of the
                # list of accepted types.
                dtype = dtype[0]

        array = NumpyMatrix(np.array(array, dtype=dtype, order=order, copy=copy))
        # make sure we acually converted to numeric:
        if dtype_numeric and array.dtype.kind == "O":
            array = array.astype(np.float64)

        return array

    def pairwise_distances(self, Y=None, metric="cosine", **kwargs):
        if Y is not None:
            Y = _check_array(Y, ensure_2d=True)
        return sklearn.metrics.pairwise_distances(self, Y, metric=metric, **kwargs)


def _check_array(array, ensure_2d=False):
    if not sparse.issparse(array):
        array = np.array(array)
        if array.ndim == 1 and ensure_2d:
            array = array.reshape(1, -1)
    return array



# -*- coding: utf-8 -*-
"""
This module provides a torch/transformers implementation of the fine-tuning procedure described in "BERT for Coreference
Resolution: Baselines and Analysis" (https://arxiv.org/pdf/1908.09091.pdf)
"""

import math

import transformers
import torch
import torch.nn.functional as F
import torch.nn as nn


class BertForCoreferenceResolutionBase(transformers.BertPreTrainedModel):
    """
    Class for fine-tuning a BERT model for the coreference resolution task.
    This is an implementation of https://arxiv.org/pdf/1908.09091.pdf, which uses the fine tuning procedure described in
    https://arxiv.org/pdf/1804.05392.pdf
    """
    def __init__(self, base_bert_config, max_span_width=30, ffnn_hidden_size=1000, top_span_ratio=0.4,
                 max_top_antecendents=50, use_metadata=False, metadata_feature_size=20,
                 genres=("bc", "bn", "mz", "nw", "pt", "tc", "wb"), max_training_segments=5, coref_depth=2,
                 coref_dropout=0.3):
        if base_bert_config.task_specific_params is None:
            # used for saving/loading model
            base_bert_config.task_specific_params = {"max_span_width": max_span_width,
                                                     "ffnn_hidden_size": ffnn_hidden_size,
                                                     "top_span_ratio": top_span_ratio,
                                                     "max_top_antecendents": max_top_antecendents,
                                                     "use_metadata": use_metadata,
                                                     "metadata_feature_size": metadata_feature_size, "genres": genres,
                                                     "max_training_segments": max_training_segments,
                                                     "coref_depth": coref_depth,
                                                     "coref_dropout": coref_dropout}

        else:
            # use config params if available
            max_span_width = base_bert_config.task_specific_params["max_span_width"] \
                if "max_span_width" in base_bert_config.task_specific_params else max_span_width
            ffnn_hidden_size = base_bert_config.task_specific_params["ffnn_hidden_size"] \
                if "ffnn_hidden_size" in base_bert_config.task_specific_params else ffnn_hidden_size
            top_span_ratio = base_bert_config.task_specific_params["top_span_ratio"] \
                if "top_span_ratio" in base_bert_config.task_specific_params else top_span_ratio
            max_top_antecendents = base_bert_config.task_specific_params["max_top_antecendents"] \
                if "max_top_antecendents" in base_bert_config.task_specific_params else max_top_antecendents
            use_metadata = base_bert_config.task_specific_params["use_metadata"] \
                if "use_metadata" in base_bert_config.task_specific_params else use_metadata
            metadata_feature_size = base_bert_config.task_specific_params["metadata_feature_size"] \
                if "metadata_feature_size" in base_bert_config.task_specific_params else metadata_feature_size
            genres = base_bert_config.task_specific_params["genres"] \
                if "genres" in base_bert_config.task_specific_params else genres
            max_training_segments = base_bert_config.task_specific_params["max_training_segments"] \
                if "max_training_segments" in base_bert_config.task_specific_params else max_training_segments
            coref_depth = base_bert_config.task_specific_params["coref_depth"] \
                if "coref_depth" in base_bert_config.task_specific_params else coref_depth
            coref_dropout = base_bert_config.task_specific_params["coref_dropout"] \
                if "coref_dropout" in base_bert_config.task_specific_params else coref_dropout
        super().__init__(base_bert_config)
        self.bert = transformers.BertModel(base_bert_config, add_pooling_layer=False)
        bert_emb_size = base_bert_config.hidden_size
        self.span_attend_projection = torch.nn.Linear(bert_emb_size, 1)
        span_embedding_dim = (bert_emb_size*3) + 20
        self.coref_dropout = coref_dropout
        self.mention_scorer = nn.Sequential(
            nn.Linear(span_embedding_dim, ffnn_hidden_size),
            nn.ReLU(),
            nn.Dropout(coref_dropout),
            nn.Linear(ffnn_hidden_size, 1),
        )
        self.width_scores = nn.Sequential(
            nn.Linear(metadata_feature_size, ffnn_hidden_size),
            nn.ReLU(),
            nn.Dropout(coref_dropout),
            nn.Linear(ffnn_hidden_size, 1),
        )
        self.fast_antecedent_projection = torch.nn.Linear(span_embedding_dim, span_embedding_dim)
        self.slow_antecedent_scorer = nn.Sequential(
            nn.Linear((span_embedding_dim * 3) + (metadata_feature_size * (4 if use_metadata else 2)),
                      ffnn_hidden_size),
            nn.ReLU(),
            nn.Dropout(coref_dropout),
            nn.Linear(ffnn_hidden_size, 1),
        )
        self.slow_antecedent_projection = torch.nn.Linear(span_embedding_dim * 2, span_embedding_dim)
        # metadata embeddings
        self.use_metadata = use_metadata
        self.genres = {g: i for i, g in enumerate(genres)}
        self.genre_embeddings = nn.Embedding(num_embeddings=len(self.genres), embedding_dim=metadata_feature_size)

        self.distance_embeddings = nn.Embedding(num_embeddings=10, embedding_dim=metadata_feature_size)
        self.slow_distance_embeddings = nn.Embedding(num_embeddings=10, embedding_dim=metadata_feature_size)
        self.distance_projection = nn.Linear(metadata_feature_size, 1)
        self.same_speaker_embeddings = nn.Embedding(num_embeddings=2, embedding_dim=metadata_feature_size)
        self.span_width_embeddings = nn.Embedding(num_embeddings=max_span_width, embedding_dim=metadata_feature_size)
        self.span_width_prior_embeddings = nn.Embedding(num_embeddings=max_span_width,
                                                        embedding_dim=metadata_feature_size)
        self.segment_dist_embeddings = nn.Embedding(num_embeddings=max_training_segments,
                                                    embedding_dim=metadata_feature_size)

        self.max_span_width = max_span_width
        self.top_span_ratio = top_span_ratio
        self.max_top_antecendents = max_top_antecendents
        self.max_training_segments = max_training_segments
        self.coref_depth = coref_depth

        self.init_weights()

    def forward(self, input_ids, attention_mask, sentence_map, speaker_ids=None, genre=None, gold_starts=None,
                gold_ends=None, cluster_ids=None, output_attentions=None, output_hidden_states=None, return_dict=None):
        """
        Pass input document through model, calculating loss if labels are present.

        Parameters
        ----------
        input_ids: tensor of size (num_segments, sequence_length)
            input token ids
        attention_mask: tensor of size (num_segments, sequence_length)
            attention mask of input segments
        sentence_map: tensor of size (num_tokens)
            sentence id for each input token in (flattened) input document
        speaker_ids: tensor of size (num_segments, sequence_length)
            speaker ids for each token (only used if self.use_metadata is True)
        genre: tensor of size (1)
            genre id for document
        gold_starts: tensor of size (labeled)
            start token indices (in flattened document) of labeled spans
        gold_ends: tensor of size (labeled)
            end token indices (in flattened document) of labeled spans
        cluster_ids: tensor of size (labeled)
            cluster ids of each labeled span
        output_attentions: Boolean
            Whether or not to return the attentions tensors of all attention layers.
        output_hidden_states: Boolean
            Whether or not to return the hidden states of all layers.
        return_dict: Boolean
            Whether or not to return a ModelOutput (dictionary) instead of a plain tuple.

        Returns
        -------
        tuple containing the following tensors if return_dict is False, else dict with following keys:
            loss:
                loss value if label input arguments (gold_starts, gold_ends, cluster_ids) are not None, else not
                returned.
            candidate_starts: tensor of size (num_spans)
                start token indices in flattened document of candidate spans
            candidate_ends: tensor of size (num_spans)
                end token indices in flattened document of candidate spans
            candidate_mention_scores: tensor of size (num_spans)
                mention scores for each candidate span
            top_span_starts: tensor of size (num_top_spans)
                start token indices in flattened document of candidate spans with top mention scores
            top_span_ends: tensor of size (num_top_spans)
                end token indices in flattened document of candidate spans with top mention scores
            top_antecedents: tensor of shape (num_top_spans, antecedent_candidates)
                indices in top span candidates of top antecedents for each mention
            top_antecedent_scores: tensor of shape (num_top_spans, antecedent_candidates)
                final antecedent scores of top antecedents for each mention
            flattened_ids: tensor of shape (num_words)
                flattened ids of input sentences. The start and end candidate indices map into this tensor.
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict
        bert_outputs = self.bert(input_ids, attention_mask, output_attentions=output_attentions,
                                 output_hidden_states=output_hidden_states, return_dict=return_dict)
        mention_doc = bert_outputs[0]   # [num_seg, max_seg_len, emb_len]
        num_seg, max_seg_len, emb_len = mention_doc.shape
        mention_doc = torch.masked_select(mention_doc.view(num_seg * max_seg_len, emb_len),
                                          attention_mask.bool().view(-1, 1)).view(-1, emb_len)  # [num_words, emb_len]
        flattened_ids = torch.masked_select(input_ids, attention_mask.bool()).view(-1)  # [num_words]
        num_words = mention_doc.shape[0]

        # calculate all possible spans
        candidate_starts = torch.arange(start=0,
                                        end=num_words,
                                        device=mention_doc.device).view(-1, 1) \
            .repeat(1, self.max_span_width)  # [num_words, max_span_width]
        candidate_ends = candidate_starts + torch.arange(start=0, end=self.max_span_width,
                                                         device=mention_doc.device).unsqueeze(
            0)  # [num_words, max_span_width]
        candidate_start_sentence_indices = sentence_map[candidate_starts]  # [num_words, max_span_width]
        candidate_end_sentence_indices = sentence_map[
            torch.clamp(candidate_ends, max=num_words - 1)]  # [num_words, max_span_width]
        # find spans that are in the same segment and don't run past the end of the text
        candidate_mask = torch.logical_and(candidate_ends < num_words,
                                           torch.eq(candidate_start_sentence_indices,
                                                    candidate_end_sentence_indices)).view(
            -1).bool()  # [num_words *max_span_width]
        candidate_starts = torch.masked_select(candidate_starts.view(-1), candidate_mask)  # [candidates]
        candidate_ends = torch.masked_select(candidate_ends.view(-1), candidate_mask)  # [candidates]

        if gold_ends is not None and gold_starts is not None and cluster_ids is not None:
            candidate_cluster_ids = self.get_candidate_labels(candidate_starts, candidate_ends, gold_starts, gold_ends,
                                                              cluster_ids)  # [candidates]

        # get span embeddings and mention scores
        span_emb = self.get_span_embeddings(mention_doc, candidate_starts, candidate_ends)  # [candidates, span_emb]
        candidate_mention_scores = self.mention_scorer(span_emb).squeeze(1)  # [candidates]
        # get span width scores and add to candidate mention scores
        span_width_index = candidate_ends - candidate_starts    # [candidates]
        span_width_emb = self.span_width_prior_embeddings(span_width_index)     # [candidates, emb]
        candidate_mention_scores += self.width_scores(span_width_emb).squeeze(1)  # [candidates]

        # get beam size
        num_top_mentions = int(float(num_words * self.top_span_ratio))
        num_top_antecedents = min(self.max_top_antecendents, num_top_mentions)

        # get top mention scores and sort by sort by span order
        top_span_indices = torch.argsort(candidate_mention_scores, descending=True)[:num_top_mentions]
        top_span_indices, _ = torch.sort(top_span_indices, descending=False)
        top_span_starts = candidate_starts[top_span_indices]  # [top_cand]
        top_span_ends = candidate_ends[top_span_indices]  # [top_cand]
        top_span_emb = span_emb[top_span_indices]  # [top_cand, span_emb]
        top_span_mention_scores = candidate_mention_scores[top_span_indices]  # [top_cand]
        if gold_ends is not None and gold_starts is not None and cluster_ids is not None:
            top_span_cluster_ids = candidate_cluster_ids[top_span_indices]  # [top_cand]
        if self.use_metadata:
            genre_emb = self.genre_embeddings(genre)  # [meta_emb]
            speaker_ids = torch.masked_select(speaker_ids.view(num_seg * max_seg_len),
                                              attention_mask.bool().view(-1))  # [top_cand]
            top_span_speaker_ids = speaker_ids[top_span_starts]  # [top_cand]
        else:
            genre_emb = None
            top_span_speaker_ids = None

        # course to fine pruning
        dummy_scores = torch.zeros([num_top_mentions, 1], device=top_span_indices.device)  # [top_cand, 1]
        top_antecedents, \
            top_antecedents_mask, \
            top_antecedents_fast_scores, \
            top_antecedent_offsets = self.coarse_to_fine_pruning(top_span_emb,
                                                                 top_span_mention_scores,
                                                                 num_top_antecedents)

        num_segments = input_ids.shape[0]
        segment_length = input_ids.shape[1]
        word_segments = torch.arange(start=0, end=num_segments, device=input_ids.device).view(-1, 1).repeat(
            [1, segment_length])  # [segments, segment_len]
        flat_word_segments = torch.masked_select(word_segments.view(-1), attention_mask.bool().view(-1))
        mention_segments = flat_word_segments[top_span_starts].view(-1, 1)  # [top_cand, 1]
        antecedent_segments = flat_word_segments[top_span_starts[top_antecedents]]  # [top_cand, top_ant]
        segment_distance = torch.clamp(mention_segments - antecedent_segments, 0,
                                       self.max_training_segments - 1)  # [top_cand, top_ant]

        # calculate final slow scores
        for i in range(self.coref_depth):
            top_antecedent_emb = top_span_emb[top_antecedents]  # [top_cand, top_ant, emb]
            top_antecedent_scores = top_antecedents_fast_scores + \
                self.get_slow_antecedent_scores(top_span_emb,
                                                top_antecedents,
                                                top_antecedent_emb,
                                                top_antecedent_offsets,
                                                top_span_speaker_ids,
                                                genre_emb,
                                                segment_distance)  # [top_cand, top_ant]
            top_antecedent_weights = F.softmax(
                torch.cat([dummy_scores, top_antecedent_scores], 1))  # [top_cand, top_ant + 1]
            top_antecedent_emb = torch.cat([top_span_emb.unsqueeze(1), top_antecedent_emb],
                                           1)  # [top_cand, top_ant + 1, emb]
            attended_span_emb = torch.sum(top_antecedent_weights.unsqueeze(2) * top_antecedent_emb,
                                          1)  # [top_cand, emb]
            gate_vectors = torch.sigmoid(
                self.slow_antecedent_projection(torch.cat([top_span_emb, attended_span_emb], 1)))  # [top_cand, emb]
            top_span_emb = gate_vectors * attended_span_emb + (1 - gate_vectors) * top_span_emb  # [top_cand, emb]

        top_antecedent_scores = torch.cat([dummy_scores, top_antecedent_scores], 1)  # [top_cand, top_ant + 1]

        # calculate loss if labels
        if gold_ends is not None and gold_starts is not None and cluster_ids is not None:
            top_antecedent_cluster_ids = top_span_cluster_ids[top_antecedents]  # [top_cand, top_ant]
            top_antecedent_cluster_ids += torch.log(top_antecedents_mask.float()).int()  # [top_cand, top_ant]
            same_cluster_indicator = torch.eq(top_antecedent_cluster_ids,
                                              top_span_cluster_ids.unsqueeze(1))  # [top_cand, top_ant]
            non_dummy_indicator = (top_span_cluster_ids > 0).unsqueeze(1)  # [top_cand, 1]
            pairwise_labels = torch.logical_and(same_cluster_indicator, non_dummy_indicator)  # [top_cand, top_ant]
            dummy_labels = torch.logical_not(pairwise_labels.any(1, keepdims=True))  # [top_cand, 1]
            top_antecedent_labels = torch.cat([dummy_labels, pairwise_labels], 1)  # [top_cand, top_ant + 1]
            loss = self.softmax_loss(top_antecedent_scores, top_antecedent_labels)  # [top_cand]
            loss = torch.sum(loss)
        else:
            loss = None

        if not return_dict:
            output = (candidate_starts, candidate_ends, candidate_mention_scores, top_span_starts, top_span_ends,
                      top_antecedents, top_antecedent_scores)
            return ((loss,) + output) if loss is not None else output

        return {
            "loss": loss,
            "candidate_starts": candidate_starts,
            "candidate_ends": candidate_ends,
            "candidate_mention_scores": candidate_mention_scores,
            "top_span_starts": top_span_starts,
            "top_span_ends": top_span_ends,
            "top_antecedents": top_antecedents,
            "top_antecedent_scores": top_antecedent_scores,
            "flattened_ids": flattened_ids
        }

    def get_slow_antecedent_scores(self, top_span_emb, top_antecedents, top_antecedent_emb, top_antecedent_offsets,
                                   top_span_speaker_ids, genre_emb, segment_distance):
        """
        Compute slow antecedent scores

        Parameters
        ----------
        top_span_emb: tensor of size (candidates, emb_size)
            span representations
        top_antecedents: tensor of size (candidates, antecedents)
            indices of antecedents for each candidate
        top_antecedent_emb: tensor of size (candidates, antecedents, emb)
            embeddings of top antecedents for each candidate
        top_antecedent_offsets: tensor of size (candidates, antecedents)
            offsets for each mention/antecedent pair
        top_span_speaker_ids: tensor of size (candidates)
            speaker ids for each span
        genre_emb: tensor of size (feature_size)
            genre embedding for document
        segment_distance: tensor of size (candidates, antecedents)
            segment distances for each candidate antecedent pair

        Returns
        -------
        tensor of shape (candidates, antecedents)
            antecedent scores
        """
        num_cand, num_ant = top_antecedents.shape
        feature_emb_list = []

        if self.use_metadata:
            top_antecedent_speaker_ids = top_span_speaker_ids[top_antecedents]  # [top_cand, top_ant]
            same_speaker = torch.eq(top_span_speaker_ids.view(-1, 1), top_antecedent_speaker_ids)  # [top_cand, top_ant]
            speaker_pair_emb = self.same_speaker_embeddings(
                torch.arange(start=0, end=2, device=top_span_emb.device))  # [2, feature_size]
            feature_emb_list.append(speaker_pair_emb[same_speaker.long()])
            genre_embs = genre_emb.view(1, 1, -1).repeat(num_cand, num_ant, 1)  # [top_cand, top_ant, feature_size]
            feature_emb_list.append(genre_embs)

        # span distance
        antecedent_distance_buckets = self.bucket_distance(top_antecedent_offsets).to(
            top_span_emb.device)  # [cand, cand]
        bucket_embeddings = self.slow_distance_embeddings(
            torch.arange(start=0, end=10, device=top_span_emb.device))  # [10, feature_size]
        feature_emb_list.append(bucket_embeddings[antecedent_distance_buckets])  # [cand, ant, feature_size]

        # segment distance
        segment_distance_emb = self.segment_dist_embeddings(
            torch.arange(start=0, end=self.max_training_segments, device=top_span_emb.device))
        feature_emb_list.append(segment_distance_emb[segment_distance])  # [cand, ant, feature_size]

        feature_emb = torch.cat(feature_emb_list, 2)  # [cand, ant, emb]
        feature_emb = F.dropout(feature_emb, p=self.coref_dropout,
                                training=self.training)  # [cand, ant, emb]
        target_emb = top_span_emb.unsqueeze(1)  # [cand, 1, emb]
        similarity_emb = top_antecedent_emb * target_emb  # [cand, ant, emb]
        target_emb = target_emb.repeat(1, num_ant, 1)  # [cand, ant, emb]

        pair_emb = torch.cat([target_emb, top_antecedent_emb, similarity_emb, feature_emb], 2)  # [cand, ant, emb]

        return self.slow_antecedent_scorer(pair_emb).squeeze(-1)

    def coarse_to_fine_pruning(self, span_emb, mention_scores, num_top_antecedents):
        """
        Compute fast estimate antecedent scores and prune based on these scores.

        Parameters
        ----------
        span_emb: tensor of size (candidates, emb_size)
            span representations
        mention_scores: tensor of size (candidates)
            mention scores of spans
        num_top_antecedents: int
            number of antecedents

        Returns
        -------
        top_antecedents: tensor of shape (mentions, antecedent_candidates)
            indices of top antecedents for each mention
        top_antecedents_mask: tensor of shape (mentions, antecedent_candidates)
            boolean mask for antecedent candidates
        top_antecedents_fast_scores: tensor of shape (mentions, antecedent_candidates)
            fast scores for each antecedent candidate
        top_antecedent_offsets: tensor of shape (mentions, antecedent_candidates)
            offsets for each mention/antecedent pair
        """
        num_candidates = span_emb.shape[0]
        top_span_range = torch.arange(start=0, end=num_candidates, device=span_emb.device)
        antecedent_offsets = top_span_range.unsqueeze(1) - top_span_range.unsqueeze(0)  # [cand, cand]
        antecedents_mask = antecedent_offsets >= 1  # [cand, cand]
        fast_antecedent_scores = mention_scores.unsqueeze(1) + mention_scores.unsqueeze(0)  # [cand, cand]
        fast_antecedent_scores += torch.log(antecedents_mask.float())  # [cand, cand]
        fast_antecedent_scores += self.get_fast_antecedent_scores(span_emb)  # [cand, cand]
        # add distance scores
        antecedent_distance_buckets = self.bucket_distance(antecedent_offsets).to(span_emb.device)  # [cand, cand]
        bucket_embeddings = F.dropout(self.distance_embeddings(torch.arange(start=0, end=10, device=span_emb.device)),
                                      p=self.coref_dropout, training=self.training)  # [10, feature_size]
        bucket_scores = self.distance_projection(bucket_embeddings)  # [10, 1]
        fast_antecedent_scores += bucket_scores[antecedent_distance_buckets].squeeze(-1)  # [cand, cand]
        # get top antecedent scores/features
        _, top_antecedents = torch.topk(fast_antecedent_scores, num_top_antecedents, sorted=False,
                                        dim=1)  # [cand, num_ant]
        top_antecedents_mask = self.batch_gather(antecedents_mask, top_antecedents)  # [cand, num_ant]
        top_antecedents_fast_scores = self.batch_gather(fast_antecedent_scores, top_antecedents)  # [cand, num_ant]
        top_antecedents_offsets = self.batch_gather(antecedent_offsets, top_antecedents)  # [cand, num_ant]
        return top_antecedents, top_antecedents_mask, top_antecedents_fast_scores, top_antecedents_offsets

    @staticmethod
    def batch_gather(emb, indices):
        batch_size, seq_len = emb.shape
        flattened_emb = emb.view(-1, 1)
        offset = (torch.arange(start=0, end=batch_size, device=indices.device) * seq_len).unsqueeze(1)
        return flattened_emb[indices + offset].squeeze(2)

    @staticmethod
    def get_candidate_labels(candidate_starts, candidate_ends, labeled_starts, labeled_ends, labels):
        """
        get labels of candidates from gold ground truth

        Parameters
        ----------
        candidate_starts, candidate_ends: tensor of size (candidates)
            start and end token indices (in flattened document) of candidate spans
        labeled_starts, labeled_ends: tensor of size (labeled)
            start and end token indices (in flattened document) of labeled spans
        labels: tensor of size (labeled)
            cluster ids

        Returns
        -------
        candidate_labels: tensor of size (candidates)
        """
        same_start = torch.eq(labeled_starts.unsqueeze(1),
                              candidate_starts.unsqueeze(0))  # [num_labeled, num_candidates]
        same_end = torch.eq(labeled_ends.unsqueeze(1), candidate_ends.unsqueeze(0))  # [num_labeled, num_candidates]
        same_span = torch.logical_and(same_start, same_end)  # [num_labeled, num_candidates]
        # type casting in next line is due to torch not supporting matrix multiplication for Long tensors
        candidate_labels = torch.mm(labels.unsqueeze(0).float(), same_span.float()).long()  # [1, num_candidates]
        return candidate_labels.squeeze(0)  # [num_candidates]

    @staticmethod
    def bucket_distance(distances):
        """
        Places the given values (designed for distances) into 10 semi-logscale buckets:
        [0, 1, 2, 3, 4, 5-7, 8-15, 16-31, 32-63, 64+].

        Parameters
        ----------
        distances: tensor of size (candidates, candidates)
            token distances between pairs

        Returns
        -------
        distance buckets
            tensor of size (candidates, candidates)
        """
        logspace_idx = torch.floor(torch.log(distances.float()) / math.log(2)).int() + 3
        use_identity = (distances <= 4).int()
        combined_idx = use_identity * distances + (1 - use_identity) * logspace_idx
        return torch.clamp(combined_idx, 0, 9)

    def get_fast_antecedent_scores(self, span_emb):
        """
        Obtains representations of the spans

        Parameters
        ----------
        span_emb: tensor of size (candidates, emb_size)
            span representations

        Returns
        -------
        fast antecedent scores
            tensor of size (candidates, span_embedding_size)
        """
        source_emb = F.dropout(self.fast_antecedent_projection(span_emb),
                               p=self.coref_dropout, training=self.training)  # [cand, emb]
        target_emb = F.dropout(span_emb, p=self.coref_dropout, training=self.training)  # [cand, emb]
        return torch.mm(source_emb, target_emb.t())  # [cand, cand]

    def get_span_embeddings(self, hidden_states, span_starts, span_ends):
        """
        Obtains representations of the spans

        Parameters
        ----------
        hidden_states: tensor of size (num_tokens, emb_size)
            outputs of BERT model, reshaped
        span_starts, span_ends: tensor of size (num_candidates)
            indices of starts and ends of spans

        Returns
        -------
        tensor of size (num_candidates, span_embedding_size)
        """
        emb = [hidden_states[span_starts], hidden_states[span_ends]]

        span_width = 1 + span_ends - span_starts  # [num_cand]
        span_width_index = span_width - 1  # [num_cand]
        span_width_emb = self.span_width_embeddings(span_width_index)  # [num_cand, emb]
        span_width_emb = F.dropout(span_width_emb, p=self.coref_dropout, training=self.training)
        emb.append(span_width_emb)

        token_attention_scores = self.get_span_word_attention_scores(hidden_states, span_starts,
                                                                     span_ends)  # [num_cand, num_words]
        attended_word_representations = torch.mm(token_attention_scores, hidden_states)  # [num_cand, emb_size]
        emb.append(attended_word_representations)
        return torch.cat(emb, dim=1)

    def get_span_word_attention_scores(self, hidden_states, span_starts, span_ends):
        """

        Parameters
        ----------
        hidden_states: tensor of size (num_tokens, emb_size)
            outputs of BERT model, reshaped
        span_starts, span_ends: tensor of size (num_candidates)
            indices of starts and ends of spans

        Returns
        -------
        tensor of size (num_candidates, span_embedding_size)
        """
        document_range = torch.arange(start=0, end=hidden_states.shape[0], device=hidden_states.device).unsqueeze(
            0).repeat(span_starts.shape[0], 1)  # [num_cand, num_words]
        token_mask = torch.logical_and(document_range >= span_starts.unsqueeze(1),
                                       document_range <= span_ends.unsqueeze(1))  # [num_cand, num_words]
        token_atten = self.span_attend_projection(hidden_states).squeeze(1).unsqueeze(0)  # [1, num_words]
        token_attn = F.softmax(torch.log(token_mask.float()) + token_atten, 1)  # [num_cand, num_words]span
        return token_attn

    @staticmethod
    def softmax_loss(top_antecedent_scores, top_antecedent_labels):
        """
        Calculate softmax loss

        Parameters
        ----------
        top_antecedent_scores: tensor of size [top_cand, top_ant + 1]
            scores of each antecedent for each mention candidate
        top_antecedent_labels: tensor of size [top_cand, top_ant + 1]
            labels for each antecedent

        Returns
        -------
        tensor of size (num_candidates)
            loss for each mention
        """
        gold_scores = top_antecedent_scores + torch.log(top_antecedent_labels.float())  # [top_cand, top_ant+1]
        marginalized_gold_scores = torch.logsumexp(gold_scores, 1)  # [top_cand]
        log_norm = torch.logsumexp(top_antecedent_scores, 1)  # [top_cand]
        return log_norm - marginalized_gold_scores  # [top_cand]



# -*- coding: utf-8 -*-
"""Classes and functions to evaluate embeddings according to the "Analogy" task.

The Analogy task tries to predict the answer of the question of the form : a is to b as c is to ...
It uses both 3CosAdd [2]_ and 3CosMul [3]_ methods to solve them

Datasets available in this module :

* `GOOGLE` for the Mikolov et al.'s (2013) Google dataset [1]_ . Also partitionned into :

    * `GOOGLE_SEMANTIC` for semantic analogies
    * `GOOGLE_SYNTACTIC` for syntactic analogies

* `MSR` for the Mikolov et al.'s (2013) Microsoft Research dataset [2]_

References
----------
.. [1] Mikolov, T., Chen, K., Corrado, G., & Dean, J. (2013). Efficient estimation of word representations in vector
       space. arXiv preprint arXiv:1301.3781.
.. [2] Mikolov, T., Yih, W. T., & Zweig, G. (2013, June). Linguistic regularities in continuous space word
       representations. In hlt-Naacl (Vol. 13, pp. 746-751).
.. [3] Levy, O., Goldberg, Y., & Ramat-Gan, I. (2014). Linguistic Regularities in Sparse and Explicit Word
       Representations. In CoNLL (pp. 171–180).


"""
import heapq
from collections import namedtuple

import numpy as np

import mangoes
import mangoes.utils.arrays
import mangoes.utils.metrics
from mangoes.evaluation.base import PrintableReport, BaseEvaluation, BaseEvaluator

_Analogy = namedtuple("Analogy", "abc gold")
_Analogy.__doc__ = """Analogy with expected answer
        
        Parameters
        ----------
        abc: str
            The 3 terms of the analogy (a is to b as c is to ...) as a single splittable string. 
        gold: str
            The expected answer
            
        Examples
        --------
        >>> Analogy('paris france london', 'england')
        Analogy(abc='paris france london', gold='england')
        >>> Dataset.parse_question('paris france london england')
        Analogy(abc='paris france london', gold='england')
        """


# ####################################################################################################################
# DATASETS

class Dataset(mangoes.evaluation.base.BaseDataset):
    """Class to create a Dataset of analogies, to be used in Evaluation class

    Examples
    --------
    >>> from mangoes.evaluation.analogy import Dataset
    >>> user_dataset = Dataset("user dataset", ['paris france london england', 'get gets do does'])
    >>> capitals = Dataset("google", "../resources/en/analogy/google/semantic/capital-world.txt")

    2 analogy datasets are available in this module:

    - the GOOGLE dataset, also split in GOOGLE_SEMANTIC and GOOGLE_SYNTACTIC :
    >>> import mangoes.evaluation.analogy
    >>> google = mangoes.evaluation.analogy.GOOGLE
    >>> google_sem = mangoes.evaluation.analogy.GOOGLE_SEMANTIC
    >>> google_syn = mangoes.evaluation.analogy.GOOGLE_SYNTACTIC

    - the MSR dataset :
    >>> import mangoes.evaluation.analogy
    >>> msr = mangoes.evaluation.analogy.MSR

    """

    @classmethod
    def parse_question(cls, question):
        """
        Examples
        --------
        >>> Dataset.parse_question('paris france london england')
        Analogy(abc='paris france london', gold='england')

        Parameters
        ----------
        question: str
            A splittable string with the 4 terms of the analogies

        Returns
        -------
        namedtuple

        """
        if isinstance(question, _Analogy):
            return question
        a, b, c, d = question.strip().split()
        return _Analogy(' '.join((a, b, c)), d)


GOOGLE = Dataset("Google", "../resources/en/analogy/google")
GOOGLE_SEMANTIC = Dataset("Google Semantic", "../resources/en/analogy/google/semantic")
GOOGLE_SYNTACTIC = Dataset("Google Syntactic", "../resources/en/analogy/google/syntactic")
MSR = Dataset("MSR", "../resources/en/analogy/msr")

ALL_DATASETS = (GOOGLE, MSR)


# ####################################################################################################################
# EVALUATOR

class Evaluator(BaseEvaluator):
    _Prediction = namedtuple("Prediction", "using_cosadd using_cosmul")

    def __init__(self, representation, threshold=300000):
        """Evaluator to resolve analogies according to the given representation

        Parameters
        ----------
        representation: mangoes.Representation
            The Representation to use
        threshold: int
            A threshold to reduce the size of vocabulary of the representation for fast approximate evaluation
            (default is 300000 as in word2vec)
        """
        self.threshold = min(threshold, len(representation.words))

        self.representation = representation
        self.vocabulary = representation.words[:threshold]

        if self.representation.matrix.all_positive():
            self.cosine_similarity = mangoes.utils.metrics.pairwise_cosine_similarity
        else:
            self.cosine_similarity = mangoes.utils.metrics.pairwise_non_negative_cosine_similarity

    def _prepare(self, abc_vocabulary):
        abc_matrix = self.representation.matrix[[self.representation.words.index(word) for word in abc_vocabulary]]
        abc_matrix = mangoes.utils.arrays.Matrix.factory(abc_matrix)

        return self.cosine_similarity(abc_matrix, self.representation.matrix[:self.threshold])

    def predict(self, analogies, allowed_answers=1, epsilon=0.001, batch=1000):
        """Predict the answer for the given analogy question(s).

        Examples
        --------
        >>> # create a representation
        >>> import numpy as np
        >>> import mangoes
        >>> vocabulary = mangoes.Vocabulary(['paris', 'france', 'london', 'england', 'belgium', 'germany'])
        >>> matrix = np.array([[1, 0], [1, 0.2], [0, 1], [0, 1.2], [0.7, 0.7], [0.7, 0.8]])
        >>> representation = mangoes.Embeddings(vocabulary, matrix)
        >>> # predict
        >>> import mangoes.evaluation.analogy
        >>> evaluator = mangoes.evaluation.analogy.Evaluator(representation)
        >>> evaluator.predict('paris france london')
        Prediction(using_cosadd=['england'], using_cosmul=['england'])

        Parameters
        ----------
        analogies: str or list of str
            an analogy or a list of analogies to resolve in the form 'a b c' : a is to b as c is to ...
        allowed_answers
            number of answers to predict
        epsilon
            value to use as epsilon when computing 3CosMul
        batch
            As this function needs to compute the similarities between all the words in the analogies and all the
            words in the vocabulary, it can be memory-consuming. This parameter allowed to slice the list in batches.
            You can increase it to run faster or decrease it if you run out of memory.

        Returns
        -------
        namedtuple or dict
            If the input is a single analogy, returns a tuple with both predictions using cosadd and cosmul.
            If the input is a list of analogies, returns a dictionary with analogies as keys and
            the predictions as values.

        """
        if not analogies:
            return {}
        if isinstance(analogies, str):
            return self._predict([analogies], allowed_answers, epsilon)[analogies]

        analogies = list(analogies)
        start = 0
        result = {}

        while start < len(analogies):
            batch_analogies = analogies[start:min(start + batch, len(analogies))]
            result.update(self._predict(batch_analogies, allowed_answers, epsilon))
            start += batch

        return result

    def _predict(self, analogies, allowed_answers, epsilon):
        # get the words in the analogies and compute the similarities between them and all the words in the vocabulary
        analogies_terms = {w for q in analogies for w in q.split()}
        analogies_terms = mangoes.Vocabulary([w for w in analogies_terms if w in self.vocabulary])
        similarities = self._prepare(analogies_terms)

        analogies_indices = np.array([[analogies_terms.index(w) for w in analogy.split()] for analogy in analogies])

        # to remove the terms of the analogies from the result, we'll affect them -inf score
        terms_indices_in_candidates_vocabulary = [[self.vocabulary.index(word)
                                                   for word in abc.split() if word in self.vocabulary]
                                                  for abc in analogies]

        # compute scores with cosadd
        scores = self._3cosadd(similarities[analogies_indices[:, 0]],
                               similarities[analogies_indices[:, 1]],
                               similarities[analogies_indices[:, 2]])
        # exclude the terms of the questions from possible answers
        for i, qti in enumerate(terms_indices_in_candidates_vocabulary):
            scores[i, qti] = np.NINF

        best_answer_add = [[self.vocabulary[i] for i in _get_n_best(allowed_answers, scores[j])]
                           for j in range(len(analogies))]

        # compute scores with cosmul
        scores = self._3cosmul(similarities[analogies_indices[:, 0]],
                               similarities[analogies_indices[:, 1]],
                               similarities[analogies_indices[:, 2]],
                               epsilon)
        for i, qti in enumerate(terms_indices_in_candidates_vocabulary):
            scores[i, qti] = np.NINF
        best_answer_mul = [[self.vocabulary[i] for i in _get_n_best(allowed_answers, scores[j])]
                           for j in range(len(analogies))]

        return {abc: self._Prediction(cosadd, cosmul)
                for abc, cosadd, cosmul in zip(analogies, best_answer_add, best_answer_mul)}

    @staticmethod
    def _3cosadd(sim_with_a, sim_with_b, sim_with_c):
        return sim_with_b - sim_with_a + sim_with_c

    @staticmethod
    def _3cosmul(sim_with_a, sim_with_b, sim_with_c, epsilon=0.001):
        return sim_with_b * sim_with_c / (sim_with_a + epsilon)


# ####################################################################################################################
# EVALUATOR

class Evaluation(BaseEvaluation):
    """Class to evaluate a representation on a dataset or a list of datasets

    Examples
    --------
    >>> # create a representation
    >>> import numpy as np
    >>> import mangoes
    >>> vocabulary = mangoes.Vocabulary(['paris', 'france', 'london', 'england', 'berlin', 'germany'])
    >>> matrix = np.array([[1, 0], [1, 0.2], [0, 1], [0, 1.2], [0.7, 0.7], [0.7, 0.8]])
    >>> representation = mangoes.Embeddings(vocabulary, matrix)
    >>> # evaluate
    >>> import mangoes.evaluation.analogy
    >>> dataset = Dataset("test", ['paris france london england', 'paris france berlin germany'])
    >>> evaluation = mangoes.evaluation.analogy.Evaluation(representation, dataset)
    >>> evaluation.get_score()
    Score(cosadd=1.0, cosmul=0.5, nb=2)
    >>> print(evaluation.get_report()) # doctest: +NORMALIZE_WHITESPACE
                                                                Nb questions      cosadd      cosmul
    ================================================================================================
    test                                                                 2/2     100.00%      50.00%
    ------------------------------------------------------------------------------------------------


    Parameters
    ----------
    representation: mangoes.Representation
        The representation to evaluate
    datasets: Dataset
        The dataset(s) to use
    lower: bool
        Whether or not the analogies in the dataset should be lowered
    allowed_answers: int
        Nb of answers to consider when predicting an analogy (the analogy will be considered as correct if the
        expected answer is among the `allowed_answers` best answers)
    epsilon: float
        Value to be used as epsilon when computing 3CosMul
    threshold: int
        A threshold to reduce the size of vocabulary of the representation for fast approximate evaluation
        (default is 300000 as in word2vec)
    """

    class _Score:
        def __init__(self, nb_correct_cosadd, nb_correct_cosmul, nb_evaluated):
            self.nb = nb_evaluated

            if nb_evaluated:
                self.cosadd = nb_correct_cosadd / nb_evaluated
                self.cosmul = nb_correct_cosmul / nb_evaluated
            else:
                self.cosadd = np.nan
                self.cosmul = np.nan

        def __repr__(self):
            return "Score(cosadd={}, cosmul={}, nb={})".format(self.cosadd, self.cosmul, self.nb)

    class _Report(PrintableReport):
        HEADER = [(("Nb questions", "cosadd", "cosmul"), (3, 3, 3))]
        PREDICTION_HEADER = []

        def _print_score(self, score):
            if score.nb:
                return "{:>{width}}{:>{width}}".format("{:.2%}".format(score.cosadd), "{:.2%}".format(score.cosmul),
                                                       width=3 * self.COL)
            else:
                return "{:>{width}}{:>{width}}".format('NA', 'NA', width=3 * self.COL)

        def _print_prediction(self, analogy, indent):
            cosadd, cosmul = self.evaluation.predictions[analogy.abc]
            line = "{:{width}}".format('    ' * indent + ' '.join(analogy), width=(self.LINE_LENGTH - 6 * self.COL))
            line += "{:>{width}}".format(", ".join(cosadd), width=3 * self.COL)
            line += "{:>{width}}".format(", ".join(cosmul), width=3 * self.COL)
            return line + "\n"

    def __init__(self, representation, *datasets, lower=True, allowed_answers=1, epsilon=0.001, threshold=30000):
        super(Evaluation, self).__init__(Evaluator(representation, threshold=threshold),
                                         *datasets, lower=lower, evaluator_kwargs={'allowed_answers': allowed_answers,
                                                                                   'epsilon': epsilon})

    def _filter_list_of_questions(self, list_of_questions):
        subset_analogies = []  # analogies that can be predicted, respecting duplicates and order from original
        unique_abc = set()  # unique analogies to resolve
        unique_abcd = set()  # unique analogies with expected answers to detect duplicates
        nb_oov = 0  # number of ignored analogies due to OOV terms
        nb_duplicates = 0  # number of duplicates among the filtered analogies

        for analogy in list_of_questions:
            analogy = _Analogy(analogy.abc.lower(), analogy.gold.lower()) if self.lower else analogy
            (a, b, c), d = analogy.abc.split(), analogy.gold

            if all(w in self.evaluator.vocabulary for w in [a, b, c, d]):
                subset_analogies.append(analogy)

                if analogy in unique_abcd:
                    nb_duplicates += 1
                else:
                    unique_abcd.add(analogy)
                unique_abc.add(analogy.abc)
            else:
                nb_oov += 1

        result = self._FilteredSubset(subset_analogies, len(list_of_questions), nb_oov, nb_duplicates)
        return result, unique_abc, unique_abcd

    def _score(self, analogies):
        nb_correct_cosadd = sum([analogy.gold in self.predictions[analogy.abc].using_cosadd for analogy in analogies])
        nb_correct_cosmul = sum([analogy.gold in self.predictions[analogy.abc].using_cosmul for analogy in analogies])
        return nb_correct_cosadd, nb_correct_cosmul


# ####################################################################################################################
# HELPER FUNCTIONS
def _get_n_best(nb_best, scores):
    """From a vector of scores, returns the indices of the n best ones"""
    if nb_best == 1:
        return [np.nanargmax(scores)]

    return heapq.nlargest(nb_best, range(len(scores)), key=lambda i: scores[i])



import collections
import random
import numpy as np
from scipy import sparse


from mangoes.utils.options import ProgressBar

Token = collections.namedtuple('Token', ("form", "lemma", "POS"), module='mangoes.utils.counting')

cdef list ngrams(list sentence, int n):
    cdef int i, j
    cdef int sentence_len = len(sentence)
    cdef num_ngrams = sentence_len - n + 1
    cdef list ngrams_results = [_ for _ in range(num_ngrams)]

    for i in range(num_ngrams):
        ngrams_results[i] = tuple([sentence[j] for j in range(i, i + n)])
    return ngrams_results

cpdef (int, int) _get_random_size(int max_size_before, int max_size_after):
    cdef int size_before, size_after
    # TODO: replace python randint with cython equivalent for speed up
    size_before = random.randint(1, max_size_before) if max_size_before > 0 else 0
    if max_size_after == max_size_before:
        size_after = size_before
    else:
        size_after = random.randint(1, max_size_after) if max_size_after > 0 else 0
    return size_before, size_after

def _counter_to_csr(counter, (int, int) shape):
    """Build a sparse.csr_matrix from a collection.Counter built with count_cooccurrence.

    Parameters
    -----------
    counter: Counter
        Counter of ((i,j), count) key-values pairs
    shape: tuple
        shape of the resulting scr matrix

    Returns
    --------
    sparse.csr_matrix
    """
    cdef Py_ssize_t counter_len = len(counter)
    cdef int i

    data = np.empty(shape=(counter_len, 3), dtype=np.intc)
    cdef int[:, ::1] data_view = data
    for i, ((word_index, context_index), count) in enumerate(counter.items()):
        data_view[i, 0] = word_index
        data_view[i, 1] = context_index
        data_view[i, 2] = count
    return sparse.csr_matrix((data_view[:, 2], (data_view[:, 0], data_view[:, 1])), shape=shape)


def _reduce_counter(tuple total, tuple part):
    """ Reduces multiprocessing counters

    Parameters
    -----------
    total: tuple
        tuple containing current total counter in scipy scr sparse matrix, and total vocabulary
    part: tuple
        tuple containing new counter in scipy scr sparse matrix, and new vocabulary.
        Will be added to total

    Returns
    --------
    tuple of form (sparse.csr_matrix, Vocabulary) containing combined counts and vocabulary
    """
    total_counter, total_vocabulary = total
    part_counter, part_vocabulary = part

    if total_vocabulary == part_vocabulary:
        return total_counter + part_counter, total_vocabulary

    cdef dict part_to_total_indices_map = {}
    cdef int i

    for i in range(len(part_vocabulary)):
        part_to_total_indices_map[i] = total_vocabulary.index(part_vocabulary[i])

    # during the mapping, words of part_vocabulary are added to total_vocabulary
    new_shape = (total_counter.shape[0], len(total_vocabulary))

    # update the indices in part_counter to map them to total_counter
    cdef int num_new_indices = part_counter.indices.shape[0]
    new_indices = np.empty(shape=(num_new_indices,), dtype=np.intc)
    cdef int[::1] new_indices_view = new_indices
    cdef int[:] part_indices_view = part_counter.indices

    for i in range(num_new_indices):
        new_indices_view[i] = part_to_total_indices_map[part_indices_view[i]]

    new_part = sparse.csr_matrix((part_counter.data, new_indices, part_counter.indptr), shape=new_shape)
    total_counter.resize(new_shape)

    return total_counter + new_part, total_vocabulary


def count_words_raw(sentences, nb_sentences=None):
    cdef int real_nb_sentences = 0
    cdef list sentence
    cdef dict words_count = {}

    for sentence in ProgressBar(sentences, total=nb_sentences, desc="Counting words"):
        for word in sentence:
            try:
                words_count[word] += 1
            except KeyError:
                words_count[word] = 1
        real_nb_sentences += 1

    return collections.Counter(words_count), real_nb_sentences


def count_words_annotated(sentences, nb_sentences=None):
    cdef int real_nb_sentences = 0
    cdef list sentence
    cdef dict words_count = {}
    cdef tuple token

    for sentence in ProgressBar(sentences, total=nb_sentences, desc="Counting words"):
        for tok in sentence:
            token = tuple((tok.form, tok.lemma, tok.POS))
            try:
                words_count[token] += 1
            except KeyError:
                words_count[token] = 1
        real_nb_sentences += 1

    return collections.Counter({Token(k[0], k[1], k[2]):v for k,v in words_count.items()}), real_nb_sentences


def count_bigrams_raw(sentences, nb_sentences=None):
    cdef int real_nb_sentences = 0
    cdef list sentence
    cdef dict bigrams_count = {}
    cdef int i, n
    cdef tuple bigram

    for sentence in ProgressBar(sentences, total=nb_sentences, desc="Counting words"):
        n = len(sentence)
        if n > 1:
            for i in range(n-1):
                bigram = tuple((sentence[i], sentence[i+1]))
                try:
                    bigrams_count[bigram] += 1
                except KeyError:
                    bigrams_count[bigram] = 1
        real_nb_sentences += 1

    return collections.Counter(bigrams_count), real_nb_sentences


def count_bigrams_annotated(sentences, nb_sentences=None):
    cdef int real_nb_sentences = 0
    cdef list sentence
    cdef dict bigrams_count = {}
    cdef tuple token, bigram
    cdef int i, n

    for sentence in ProgressBar(sentences, total=nb_sentences, desc="Counting words"):
        n = len(sentence)
        if n > 1:
            for i in range(n-1):
                bigram = tuple((tuple((sentence[i].form, sentence[i].lemma, sentence[i].POS)),
                                tuple((sentence[i+1].form, sentence[i+1].lemma, sentence[i+1].POS))))
                try:
                    bigrams_count[bigram] += 1
                except KeyError:
                    bigrams_count[bigram] = 1
        real_nb_sentences += 1

    return collections.Counter({tuple((Token(k[0][0], k[0][1], k[0][2]), Token(k[1][0], k[1][1], k[1][2]))):v for k,v
                                in bigrams_count.items()}), real_nb_sentences


def get_window_function(int size_before, int size_after, bint dirty, bint dynamic, int n, bint distance):
    # if distance is True, the items returned by _window are tuples : (token, distance from target)
    format_output = (lambda token, d: (token, d)) if distance else (lambda token, _: token)

    # if dynamic, the size of the window is sampled between 1 and the size passed as parameter
    get_size = (lambda: _get_random_size(size_before, size_after)) if dynamic else (lambda: (size_before, size_after))

    # if n > 1, _window returns n-grams i.e. tuples of n tokens
    get_items_from = (lambda sentence: ngrams(sentence, n)) if n > 1 else (lambda sentence: sentence)

    # if n > 1, all tokens in n-grams have to be checked if in vocabulary
    ignored = {None, -1}
    check_in_vocabulary = (lambda ng: not any((t in ignored for t in ng))) if n > 1 else (lambda t: t not in ignored)

    if not dirty:
        if not dynamic and size_before == size_after and n == 1 and not distance:
            def _window(list sentence, list mask):
                cdef int sentence_len = len(sentence)
                cdef list result = [[] for _ in range(sentence_len)]
                cdef int i, j

                for i in range(sentence_len):
                    if mask[i]:
                        result[i].extend([sentence[j] for j in range(max(i - size_before, 0),
                                                                     min(i + size_after + 1, sentence_len))
                                          if not i == j and sentence[j] not in ignored])
                    else:
                        result[i] = []
                return result
        else:
            def _window(list sentence, list mask):
                cdef int sentence_len = len(sentence)
                cdef list result = [[] for _ in range(sentence_len)]
                cdef list before_list, after_list
                cdef int i, j, start_index, start_distance, end_index, before, after

                for i in range(sentence_len):
                    if mask[i]:
                        before, after = get_size()
                        if i - before < 0:
                            start_index = 0
                            start_distance = i - before + n - 1
                        else:
                            start_index = i - before
                            start_distance = - before + n - 1

                        # distance_start is the distance from the first token in before list to the target token (i)
                        before_list = [format_output(w, d)
                                       for d, w in enumerate(get_items_from(sentence[start_index:i]),
                                                             start=start_distance)
                                       if check_in_vocabulary(w)]

                        end_index = min(i + after + 1, len(sentence))
                        after_list = [format_output(w, d)
                                      for d, w in enumerate(get_items_from(sentence[i + 1:end_index]),
                                                            start=1)
                                      if check_in_vocabulary(w)]

                        result[i] = before_list + after_list
                    else:
                        result[i] = []
                return result
    else:
        def _window(list sentence, list mask):
            cdef int sentence_len = len(sentence)
            cdef list result = [[] for _ in range(sentence_len)]
            cdef list before_list, after_list
            cdef int i, j, d, before, after

            for i in range(len(sentence)):
                if mask[i]:
                    before, after = get_size()

                    before_list = []
                    d = -1
                    while i + d >= 0 and len(before_list) < before:
                        items = get_items_from(sentence[i + d + 1 - n:i + d + 1])
                        if len(items) > 0 and check_in_vocabulary(items[0]):
                            # items cannot have more than one element
                            before_list.append(format_output(items[0], d))
                        d -= 1
                    before_list = list(reversed(before_list))

                    after_list = []
                    d = 1
                    while i + d < len(sentence) and len(after_list) < after:
                        items = get_items_from(sentence[i + d:i + d + n])
                        if len(items) > 0 and check_in_vocabulary(items[0]):
                            # items cannot have more than one element
                            after_list.append(format_output(items[0], d))
                        d += 1

                    result[i] = before_list + after_list
                else:
                    result[i] = []
            return result

    return _window




# -*- coding: utf-8 -*-
"""Class and functions to add optional features (progress bars, ...).

"""
import logging

logger = logging.getLogger(__name__)


def _in_notebook():
    """
    Returns True if the module is running in a Jupyter notebook, False if in IPython shell or other Python shell.
    """
    import sys
    return 'ipykernel' in sys.modules


class _NoProgressBar:
    """Default class to use if no progress bar library is available"""

    def __init__(self, iterable=None, **kwargs):
        self.iterable = iterable

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        pass

    def __iter__(self):
        for obj in self.iterable:
            yield obj

    def update(self):
        pass


try:
    if _in_notebook():
        from tqdm import tqdm_notebook

        ProgressBar = tqdm_notebook
    else:
        from tqdm import tqdm

        ProgressBar = tqdm
except ImportError:
    logger.info("Install tqdm to add progress bars")
    ProgressBar = _NoProgressBar
